"""File includes api to uploading data into Scout"""

import logging
from pathlib import Path
from datetime import datetime

import requests
from ruamel import yaml

from cg.apps.hk import HousekeeperAPI
from cg.apps.scout.scoutapi import ScoutAPI
from cg.apps.lims import LimsAPI
from cg.apps.madeline.api import MadelineAPI
from cg.meta.workflow.mip import MipAnalysisAPI
from cg.store import models
from cg.apps.scout.scout_load_config import ScoutLoadConfig, ScoutIndividual

from housekeeper.store import models as hk_models

LOG = logging.getLogger(__name__)


class UploadScoutAPI:
    """Class that handles everything that has to do with uploading to Scout"""

    def __init__(
        self,
        hk_api: HousekeeperAPI,
        scout_api: ScoutAPI,
        lims_api: LimsAPI,
        analysis_api: MipAnalysisAPI,
        madeline_api: MadelineAPI,
    ):
        self.housekeeper = hk_api
        self.scout = scout_api
        self.madeline_api = madeline_api
        self.analysis = analysis_api
        self.lims = lims_api

    def fetch_file_path(self, tag: str, sample_id: str, hk_version_id: int = None):
        """"Fetch files from housekeeper"""
        tags = [tag, sample_id]
        hk_file = self.housekeeper.files(version=hk_version_id, tags=tags).first()
        file_path = None
        if hk_file:
            file_path = hk_file.full_path
        return file_path

    def build_samples(self, analysis_obj: models.Analysis, hk_version_id: int = None):
        """Loop over the samples in an analysis and build dicts from them"""
        LOG.info("Building samples")
        for link_obj in analysis_obj.family.links:
            LOG.debug("Found link obj %ss", link_obj)
            sample_id = link_obj.sample.internal_id
            bam_path = self.fetch_file_path("bam", sample_id, hk_version_id)
            alignment_file_path = self.fetch_file_path("cram", sample_id, hk_version_id)
            chromograph_path = self.fetch_file_path("chromograph", sample_id, hk_version_id)
            mt_bam_path = self.fetch_file_path("bam-mt", sample_id, hk_version_id)
            vcf2cytosure_path = self.fetch_file_path("vcf2cytosure", sample_id, hk_version_id)

            lims_sample = dict()
            try:
                lims_sample = self.lims.sample(sample_id) or {}
            except requests.exceptions.HTTPError as ex:
                LOG.info("Could not fetch sample %s from LIMS: %s", sample_id, ex)
            sample = {
                "analysis_type": link_obj.sample.application_version.application.analysis_type,
                "bam_path": bam_path,
                "capture_kit": None,
                "alignment_path": alignment_file_path,
                "chromograph": chromograph_path,
                "father": link_obj.father.internal_id if link_obj.father else "0",
                "mother": link_obj.mother.internal_id if link_obj.mother else "0",
                "mt_bam": mt_bam_path,
                "phenotype": link_obj.status,
                "sample_id": sample_id,
                "sample_name": link_obj.sample.name,
                "sex": link_obj.sample.sex,
                "tissue_type": lims_sample.get("source", "unknown"),
                "vcf2cytosure": vcf2cytosure_path,
            }
            scout_sample = ScoutIndividual(**sample)
            yield scout_sample.dict(exclude_none=True)

    def generate_config(
        self, analysis_obj: models.Analysis, rank_score_threshold: int = 5
    ) -> ScoutLoadConfig:
        """Fetch data about an analysis to load Scout."""
        LOG.info("Generate scout load config")
        analysis_date: datetime = analysis_obj.started_at or analysis_obj.completed_at
        LOG.debug("Found analysis date %s", analysis_date)
        # Fetch last version from housekeeper
        # This should be safe since analyses are only added if data is analysed
        hk_version: hk_models.Version = self.housekeeper.last_version(
            analysis_obj.family.internal_id
        )
        LOG.debug("Found housekeeper version %s", hk_version.id)
        analysis_data: dict = self.analysis.get_latest_metadata(analysis_obj.family.internal_id)
        genome_build = "38" if "38" in analysis_data.get("genome_build", "") else "37"
        data = {
            "analysis_date": analysis_obj.completed_at,
            "default_gene_panels": analysis_obj.family.panels,
            "family": analysis_obj.family.internal_id,
            "family_name": analysis_obj.family.name,
            "rank_score_threshold": rank_score_threshold,
            "gene_panels": self.analysis.convert_panels(
                analysis_obj.family.customer.internal_id, analysis_obj.family.panels
            )
            or None,
            "human_genome_build": genome_build,
            "owner": analysis_obj.family.customer.internal_id,
            "rank_model_version": analysis_data.get("rank_model_version"),
            "samples": list(),
            "sv_rank_model_version": analysis_data.get("sv_rank_model_version"),
        }
        for sample in self.build_samples(analysis_obj, hk_version.id):
            data["samples"].append(sample)
        self._include_mandatory_files(data, hk_version)
        self._include_peddy_files(data, hk_version)
        self._include_optional_files(data, hk_version)

        if self._is_multi_sample_case(data):
            if self._is_family_case(data):
                svg_path = self._run_madeline(analysis_obj.family)
                data["madeline"] = svg_path
            else:
                LOG.info("family of unconnected samples - skip pedigree graph")
        else:
            LOG.info("family of 1 sample - skip pedigree graph")

        # Validate that the config is correct

        scout_case = ScoutLoadConfig(**data)

        return scout_case

    @staticmethod
    def get_load_config_tag() -> str:
        """Get the hk tag for a scout load config"""
        return "scout-load-config"

    @staticmethod
    def save_config_file(upload_config: ScoutLoadConfig, file_path: Path):
        """Save a scout load config file to <file_path>"""

        LOG.info("Save Scout load config to %s", file_path)
        yml = yaml.YAML()
        yml.dump(upload_config.dict(exclude_none=True), file_path)

    def add_scout_config_to_hk(self, config_file_path: Path, case_id: str):
        """Add scout load config to hk bundle"""
        LOG.info("Adding load config to housekeeper")
        tag_name = UploadScoutAPI.get_load_config_tag()
        version_obj = self.housekeeper.last_version(bundle=case_id)
        uploaded_config_files = self.housekeeper.get_files(
            bundle=case_id, tags=[tag_name], version=version_obj.id
        )
        LOG.info("Config files: %s", uploaded_config_files)
        number_of_configs = sum(1 for i in uploaded_config_files)
        bundle_config_exists = number_of_configs > 0

        if bundle_config_exists:
            raise FileExistsError("Upload config already exists")

        file_obj = self.housekeeper.add_file(str(config_file_path), version_obj, tag_name)
        self.housekeeper.include_file(file_obj, version_obj)
        self.housekeeper.add_commit(file_obj)

        LOG.info("Added scout load config to housekeeper: %s", config_file_path)
        return file_obj

    def _include_optional_files(self, data: dict, hk_version) -> None:
        """"Optional files on case level"""
        scout_hk_map = [
            ("delivery_report", "delivery-report"),
            ("multiqc", "multiqc-html"),
            ("vcf_str", "vcf-str"),
            ("smn_tsv", "smn-calling"),
        ]
        self._include_files(data, hk_version, scout_hk_map)

    def _include_peddy_files(self, data, hk_version):
        scout_hk_map = [
            ("peddy_ped", "ped"),
            ("peddy_sex", "sex-check"),
            ("peddy_check", "ped-check"),
        ]
        self._include_files(data, hk_version, scout_hk_map, extra_tag="peddy")

    def _include_mandatory_files(self, data, hk_version):
        scout_hk_map = {
            ("vcf_snv", "vcf-snv-clinical"),
            ("vcf_snv_research", "vcf-snv-research"),
            ("vcf_sv", "vcf-sv-clinical"),
            ("vcf_sv_research", "vcf-sv-research"),
        }
        self._include_files(data, hk_version, scout_hk_map, skip_missing=False)

    def _include_files(self, data, hk_version, scout_hk_map, **kwargs):
        extra_tag = kwargs.get("extra_tag")
        skip_missing = kwargs.get("skip_missing", True)
        for scout_key, hk_tag in scout_hk_map:

            if extra_tag:
                tags = [extra_tag, hk_tag]
            else:
                tags = [hk_tag]

            hk_file = self.housekeeper.files(version=hk_version.id, tags=tags).first()
            if hk_file:
                data[scout_key] = str(hk_file.full_path)
            else:
                if skip_missing:
                    LOG.debug("skipping missing file: %s", scout_key)
                else:
                    raise FileNotFoundError(f"missing file: {scout_key}")

    @staticmethod
    def _is_family_case(data: dict):
        """Check if there are any linked individuals in a case"""
        for sample in data["samples"]:
            if sample["mother"] and sample["mother"] != "0":
                return True
            if sample["father"] and sample["father"] != "0":
                return True
        return False

    @staticmethod
    def _is_multi_sample_case(data):
        return len(data["samples"]) > 1

    def _run_madeline(self, family_obj: models.Family):
        """Generate a madeline file for an analysis."""
        samples = [
            {
                "sample": link_obj.sample.name,
                "sex": link_obj.sample.sex,
                "father": link_obj.father.name if link_obj.father else None,
                "mother": link_obj.mother.name if link_obj.mother else None,
                "status": link_obj.status,
            }
            for link_obj in family_obj.links
        ]
        svg_path = self.madeline_api.run(family_id=family_obj.name, samples=samples)
        return svg_path



"""Tests for reset part of the store API"""

from cg.store import Store
from cg.store.api.import_func import (
    prices_are_same,
    versions_are_same,
    import_application_versions,
    _get_tag_from_raw_version,
    add_version_from_raw,
    import_applications,
    _get_tag_from_raw_application,
    import_apptags,
    XlFileHelper,
    XlSheetHelper,
)


def test_prices_are_same_int_and_int():
    # GIVEN float price that looks like one in excel
    # GIVEN same price but as it looks when saved in database
    float_price = 0.0
    database_price = 0

    # WHEN calling prices are same
    should_be_same = prices_are_same(database_price, float_price)

    # THEN prices should be considered same
    assert should_be_same


def test_prices_are_not_same_int_and_int():
    # GIVEN float price that looks like one in excel
    # GIVEN same price but as it looks when saved in database
    float_price = 1.0
    database_price = 0

    # WHEN calling prices are same
    should_not_be_same = prices_are_same(database_price, float_price)

    # THEN prices should be considered same
    assert not should_not_be_same


def test_prices_are_same_float_and_int():
    # GIVEN float price that looks like one in excel
    # GIVEN same price but as it looks when saved in database
    float_price = 0.654345
    database_price = 1

    # WHEN calling prices are same
    should_be_same = prices_are_same(database_price, float_price)

    # THEN prices should be considered same
    assert should_be_same


def test_prices_are_same_float_and_float():
    # GIVEN float price that looks like one in excel
    # GIVEN same price but as it looks when saved in database
    float_price = 0.654345
    database_price = float_price

    # WHEN calling prices are same
    should_be_same = prices_are_same(database_price, float_price)

    # THEN prices should be considered same
    assert should_be_same


def test_versions_are_same(store: Store, application_versions_file):
    # GIVEN an excel price row
    # same price row committed to the database
    add_applications(store, application_versions_file)
    raw_version = XlFileHelper.get_raw_dicts_from_xl(application_versions_file)[0]
    tag = _get_tag_from_raw_version(raw_version)
    application_obj = store.application(tag)
    sign = "DummySign"
    workbook = XlFileHelper.get_workbook_from_xl(application_versions_file)
    db_version = add_version_from_raw(application_obj, None, raw_version, sign, store, workbook)
    datemode = XlFileHelper.get_datemode_from_xl(application_versions_file)

    # WHEN calling versions are same
    should_be_same = versions_are_same(db_version, raw_version, datemode)

    # THEN versions are considered same
    assert should_be_same


def test_versions_are_not_same(store, application_versions_file):
    # GIVEN an excel price row
    # NOT same price row committed to the database
    add_applications(store, application_versions_file)
    raw_version = XlFileHelper.get_raw_dicts_from_xl(application_versions_file)[0]
    tag = _get_tag_from_raw_version(raw_version)
    application_obj = store.application(tag)
    sign = "DummySign"
    workbook = XlFileHelper.get_workbook_from_xl(application_versions_file)
    db_version = add_version_from_raw(application_obj, None, raw_version, sign, store, workbook)
    datemode = XlFileHelper.get_datemode_from_xl(application_versions_file)
    another_raw_version = XlFileHelper.get_raw_dicts_from_xl(application_versions_file)[1]

    # WHEN calling versions are same
    should_not_be_same = versions_are_same(db_version, another_raw_version, datemode)

    # THEN versions are not considered same
    assert not should_not_be_same


def test_application_version(application_versions_file, store: Store):
    # GIVEN a store with applications
    # and an excel file with prices for those applications
    add_applications(store, application_versions_file)
    sign = "TestSign"

    # WHEN calling import_application_version
    import_application_versions(
        store, application_versions_file, sign, dry_run=False, skip_missing=False
    )

    # THEN versions should have been created in the store
    assert all_versions_exists_in_store(store, application_versions_file)


def test_application_version_dry_run(application_versions_file, store: Store):
    # GIVEN a store with applications
    # and an excel file with prices for those applications
    add_applications(store, application_versions_file)
    sign = "TestSign"

    # WHEN calling import_application_version as dry run
    import_application_versions(
        store, application_versions_file, sign, dry_run=True, skip_missing=False
    )

    # THEN versions should not have been created in the store
    assert not all_versions_exists_in_store(store, application_versions_file)


def test_application(applications_file, store: Store):
    # GIVEN a store and an excel file with applications
    sign = "TestSign"

    # WHEN calling import_applications
    import_applications(store, applications_file, sign, dry_run=False)

    # THEN applications should have been created in the store
    assert all_applications_exists(store, applications_file)


def test_application_sheet_name(applications_file, store: Store):
    # GIVEN a store and an excel file with applications
    sign = "TestSign"

    # WHEN calling import_applications
    import_applications(store, applications_file, sign, dry_run=False, sheet_name="application")

    # THEN applications should have been created in the store
    assert all_applications_exists(store, applications_file)


def test_application_dry_run(applications_file, store: Store):
    # GIVEN a store and an excel file with applications
    sign = "TestSign"

    # WHEN calling import_applications as dry run
    import_applications(store, applications_file, sign, dry_run=True)

    # THEN applications should not have been created in the store
    assert not all_applications_exists(store, applications_file)


def test_sync_microbial_orderform_dryrun(base_store: Store, microbial_orderform):
    # GIVEN a microbial orderform and a store where all the apptags exists half some inactive and
    # some active
    ensure_applications(
        base_store,
        ["MWRNXTR003", "MWGNXTR003", "MWMNXTR003", "MWLNXTR003"],
        ["MWXNXTR003", "VWGNXTR001", "VWLNXTR001"],
    )
    prep_category = "mic"
    sign = "PG"
    activate = False
    inactivate = False
    active_mic_apps_from_start = base_store.applications(
        category=prep_category, archived=False
    ).count()
    inactive_mic_apps_from_start = base_store.applications(
        category=prep_category, archived=True
    ).count()

    # WHEN syncing app-tags in that orderform
    import_apptags(
        store=base_store,
        excel_path=microbial_orderform,
        prep_category=prep_category,
        sign=sign,
        activate=activate,
        inactivate=inactivate,
        sheet_name="Drop down list",
        tag_column=2,
    )

    # THEN same number of active and inactive mic applications in status database
    active_mic_apps_after_when = base_store.applications(
        category=prep_category, archived=False
    ).count()
    inactive_mic_apps_after_when = base_store.applications(
        category=prep_category, archived=True
    ).count()
    assert active_mic_apps_from_start == active_mic_apps_after_when
    assert inactive_mic_apps_from_start == inactive_mic_apps_after_when


def test_sync_microbial_orderform_activate(base_store: Store, microbial_orderform):
    # GIVEN a microbial orderform and a store where all the apptags exists half some inactive and
    # some active
    ensure_applications(
        base_store,
        ["MWRNXTR003", "MWGNXTR003", "MWMNXTR003", "MWLNXTR003"],
        ["MWXNXTR003", "VWGNXTR001", "VWLNXTR001"],
    )
    prep_category = "mic"
    sign = "PG"
    activate = True
    inactivate = False
    active_mic_apps_from_start = base_store.applications(
        category=prep_category, archived=False
    ).count()
    inactive_mic_apps_from_start = base_store.applications(
        category=prep_category, archived=True
    ).count()

    # WHEN syncing app-tags in that orderform
    import_apptags(
        store=base_store,
        excel_path=microbial_orderform,
        prep_category=prep_category,
        sign=sign,
        activate=activate,
        inactivate=inactivate,
        sheet_name="Drop down list",
        tag_column=2,
    )

    # THEN more active mic applications in status database and same inactive
    active_mic_apps_after_when = base_store.applications(
        category=prep_category, archived=False
    ).count()
    inactive_mic_apps_after_when = base_store.applications(
        category=prep_category, archived=True
    ).count()
    assert active_mic_apps_from_start < active_mic_apps_after_when
    assert inactive_mic_apps_from_start > inactive_mic_apps_after_when


def test_sync_microbial_orderform_inactivate(base_store: Store, microbial_orderform):
    # GIVEN a microbial orderform and a store where all the apptags exists half some inactive and
    # some active
    ensure_applications(
        base_store,
        ["MWRNXTR003", "MWGNXTR003", "MWMNXTR003", "MWLNXTR003", "dummy_tag"],
        ["MWXNXTR003", "VWGNXTR001", "VWLNXTR001"],
    )
    prep_category = "mic"
    sign = "PG"
    activate = False
    inactivate = True
    active_mic_apps_from_start = base_store.applications(
        category=prep_category, archived=False
    ).count()
    inactive_mic_apps_from_start = base_store.applications(
        category=prep_category, archived=True
    ).count()

    # WHEN syncing app-tags in that orderform
    import_apptags(
        store=base_store,
        excel_path=microbial_orderform,
        prep_category=prep_category,
        sign=sign,
        activate=activate,
        inactivate=inactivate,
        sheet_name="Drop down list",
        tag_column=2,
    )

    # THEN same number of active and more inactive mic applications in status database
    active_mic_apps_after_when = base_store.applications(
        category=prep_category, archived=False
    ).count()
    inactive_mic_apps_after_when = base_store.applications(
        category=prep_category, archived=True
    ).count()
    assert active_mic_apps_from_start > active_mic_apps_after_when
    assert inactive_mic_apps_from_start < inactive_mic_apps_after_when


def ensure_applications(base_store: Store, active_applications: list, inactive_applications: list):
    """Create some requested applications for the tests """
    for active_application in active_applications:
        if not base_store.application(active_application):
            base_store.add_commit(
                base_store.add_application(
                    active_application,
                    "mic",
                    "dummy_description",
                    is_archived=False,
                    percent_kth=80,
                )
            )

    for inactive_application in inactive_applications:
        if not base_store.application(inactive_application):
            base_store.add_commit(
                base_store.add_application(
                    inactive_application,
                    category="mic",
                    description="dummy_description",
                    is_archived=True,
                    percent_kth=80,
                )
            )


def ensure_application(store, tag):
    """Ensure that the specified application exists in the store"""
    application = store.application(tag=tag)
    if not application:
        application = store.add_application(
            tag=tag,
            category="wgs",
            description="dummy_description",
            is_external=False,
            percent_kth=80,
        )
        store.add_commit(application)

    return application


def add_applications(store, application_versions_file):
    """Ensure all applications in the xl exists"""

    raw_versions = XlFileHelper.get_raw_dicts_from_xl(application_versions_file)

    for raw_version in raw_versions:
        tag = _get_tag_from_raw_version(raw_version)
        ensure_application(store, tag)


def get_prices_from_store(store, raw_price):
    """Gets all versions for the specified application"""
    tag = _get_tag_from_raw_version(raw_price)
    return store.application(tag).versions


def get_application_from_store(store, raw_application):
    """Gets the specified application"""
    tag = _get_tag_from_raw_application(raw_application)
    return store.application(tag)


def exists_version_in_store(raw_price, store, datemode):
    """Check if the given raw version exists in the store"""
    db_versions = get_prices_from_store(store, raw_price)

    version_found = False
    for db_price in db_versions:
        if versions_are_same(db_price, raw_price, datemode):
            version_found = True

    return version_found


def all_versions_exists_in_store(store, excel_path):
    """Check if all versions in the excel exists in the store"""
    raw_versions = XlFileHelper.get_raw_dicts_from_xl(excel_path)
    datemode = XlFileHelper.get_datemode_from_xl(excel_path)
    for raw_version in raw_versions:
        if not exists_version_in_store(raw_version, store, datemode):
            return False

    return True


def all_applications_exists(store, applications_file):
    """Check if all applications in the excel exists in the store"""
    raw_applications = XlFileHelper.get_raw_dicts_from_xl(applications_file)

    for raw_application in raw_applications:
        if not exists_application_in_store(raw_application, store):
            return False

    return True


def exists_application_in_store(raw_application, store):
    """Check if the given raw application exists in the store"""
    db_application = get_application_from_store(store, raw_application)

    return db_application is not None



"""Code for uploading to scout via CLI"""
import logging
from pathlib import Path

import click

from cg.apps.hk import HousekeeperAPI
from cg.apps.scout.scoutapi import ScoutAPI
from cg.store import Store
from cg.store.models import Family
from cg.meta.upload.scoutapi import UploadScoutAPI
from cg.apps.scout.scout_load_config import ScoutLoadConfig

from .utils import suggest_cases_to_upload

LOG = logging.getLogger(__name__)


@click.command()
@click.option("-r", "--re-upload", is_flag=True, help="re-upload existing analysis")
@click.option("-p", "--print", "print_console", is_flag=True, help="print config values")
@click.argument("case_id", required=False)
@click.pass_context
def scout(context, re_upload: bool, print_console: bool, case_id: str):
    """Upload variants from analysis to Scout."""

    LOG.info("----------------- SCOUT -----------------------")

    if not case_id:
        suggest_cases_to_upload(context)
        context.abort()

    status_api: Store = context.obj["status_db"]
    scout_upload_api: UploadScoutAPI = context.obj["scout_upload_api"]
    family_obj: Family = status_api.family(case_id)
    scout_load_config: ScoutLoadConfig = scout_upload_api.generate_config(family_obj.analyses[0])
    mip_dna_root_dir: Path = Path(context.obj["mip-rd-dna"]["root"])
    if print_console:
        click.echo(scout_load_config.dict(exclude_none=True))
        return

    file_path = mip_dna_root_dir / case_id / "scout_load.yaml"

    if file_path.exists():
        message = (
            "Scout load config %s already exists, you might remove the file and try "
            "again, consider that you might also have it in housekeeper" % file_path
        )
        LOG.warning(message)
        raise click.Abort

    scout_upload_api.save_config_file(scout_load_config, file_path)
    try:
        LOG.info("Upload file to housekeeper: %s", file_path)
        scout_upload_api.add_scout_config_to_hk(config_file_path=file_path, case_id=case_id)
    except FileExistsError as err:
        LOG.warning("%s, consider removing the file from housekeeper and try again", str(err))
        context.abort()

    context.invoke(upload_case_to_scout, case_id=case_id, re_upload=re_upload)


@click.command()
@click.option("-r", "--re-upload", is_flag=True, help="re-upload existing analysis")
@click.option("--dry-run", is_flag=True)
@click.argument("case_id")
@click.pass_context
def upload_case_to_scout(context, re_upload, dry_run, case_id):
    """Upload variants and case from analysis to Scout."""

    LOG.info("----------------- CONFIG -----------------------")

    def _get_load_config_from_hk(hk_api: HousekeeperAPI, case_id):
        tag_name = UploadScoutAPI.get_load_config_tag()
        version_obj = hk_api.last_version(case_id)
        scout_config_files = hk_api.get_files(
            bundle=case_id, tags=[tag_name], version=version_obj.id
        )
        if len(list(scout_config_files)) == 0:
            raise FileNotFoundError(f"No scout load config was found in housekeeper for {case_id}")

        return scout_config_files[0].full_path

    scout_api: ScoutAPI = context.obj["scout_api"]
    hk_api: HousekeeperAPI = context.obj["housekeeper_api"]

    load_config = _get_load_config_from_hk(hk_api, case_id)

    LOG.info("uploading case %s to scout", case_id)

    if not dry_run:
        scout_api.upload(scout_load_config=load_config, force=re_upload)

    LOG.info("uploaded to scout using load config %s", load_config)



"""Code for talking to Scout regarding uploads"""

import logging
from typing import List, Optional
from pathlib import Path
import yaml
import json

from subprocess import CalledProcessError

from cg.utils.commands import Process
from cg.apps.scout.scout_load_config import ScoutLoadConfig
from cg.apps.scout.scout_export import ScoutExportCase, Variant

LOG = logging.getLogger(__name__)


class ScoutAPI:

    """Interface to Scout."""

    def __init__(self, config):

        binary_path = config["scout"]["binary_path"]
        config_path = config["scout"]["config_path"]
        self.process = Process(binary=binary_path, config=config_path)

    def upload(self, scout_load_config: Path, threshold: int = 5, force: bool = False):
        """Load analysis of a new family into Scout."""
        with open(scout_load_config, "r") as stream:
            data = yaml.safe_load(stream)
        scout_load_config_object: ScoutLoadConfig = ScoutLoadConfig(**data)
        existing_case: Optional[ScoutExportCase] = self.get_case(
            case_id=scout_load_config_object.family
        )
        load_command = ["load", "case", str(scout_load_config)]
        if existing_case:
            if force or scout_load_config_object.analysis_date > existing_case.analysis_date:
                load_command.append("--update")
                LOG.info("update existing Scout case")
            else:
                existing_date = existing_case.analysis_date.date()
                LOG.warning("analysis of case already loaded: %s", existing_date)
                return
        LOG.debug("load new Scout case")
        self.process.run_command(load_command)
        LOG.debug("Case loaded successfully to Scout")

    def update_alignment_file(self, case_id: str, sample_id: str, alignment_path: Path):
        """Update alignment file for individual in case"""
        parameters = [
            "update",
            "individual",
            "--case-id",
            case_id,
            "--ind-id",
            sample_id,
            "--alignment-path",
            str(alignment_path),
        ]
        self.process.run_command(parameters=parameters)

    def export_panels(self, panels: List[str], build: str = None) -> List[str]:
        """Pass through to export of a list of gene panels.

        Return list of lines in bed format
        """
        export_panels_command = ["export", "panels", "--bed"]
        for panel_id in panels:
            export_panels_command.append(panel_id)

        if build:
            export_panels_command.extend(["--build", build])

        try:
            self.process.run_command(export_panels_command)
            if not self.process.stdout:
                return []
        except CalledProcessError:
            LOG.info("Could not find panels")
            return []

        return [line for line in self.process.stdout_lines()]

    def get_genes(self, panel_id: str, build: str = None) -> list:
        """Fetch panel genes.

        Args:
            panel_id (str): unique id for the panel
            version (str): version of the panel. If 'None' latest version will be returned

        Returns:
            panel genes: panel genes list
        """
        # This can be run from CLI with `scout export panels <panel1> `
        export_panel_command = ["export", "panel", panel_id]
        if build:
            export_panel_command.extend(["--build", build])

        try:
            self.process.run_command(export_panel_command)
            if not self.process.stdout:
                return []
        except CalledProcessError:
            LOG.info("Could not find panel %s", panel_id)
            return []

        panel_genes = []
        for gene_line in self.process.stdout_lines():
            if gene_line.startswith("#"):
                continue
            gene_info = gene_line.strip().split("\t")
            if not len(gene_info) > 1:
                continue
            panel_genes.append({"hgnc_id": int(gene_info[0]), "hgnc_symbol": gene_info[1]})

        return panel_genes

    def get_causative_variants(self, case_id: str) -> List[Variant]:
        """
        Get causative variants for a case
        """
        # These commands can be run with `scout export variants`
        get_causatives_command = ["export", "variants", "case-id", case_id]
        try:
            self.process.run_command(get_causatives_command)
            if not self.process.stdout:
                return []
        except CalledProcessError:
            LOG.warning("Could not find case %s in scout", case_id)
            return []
        variants: List[Variant] = []
        for variant_info in json.loads(self.process.stdout):
            variants.append(Variant(**variant_info))
        return variants

    def get_case(self, case_id: str) -> Optional[ScoutExportCase]:
        """Fetch a case from Scout"""
        cases: List[ScoutExportCase] = self.get_cases(case_id=case_id)
        if not cases:
            return None
        return cases[0]

    def get_cases(
        self,
        case_id: Optional[str] = None,
        reruns: bool = False,
        finished: bool = False,
        status: Optional[str] = None,
        days_ago: int = None,
    ) -> List[ScoutExportCase]:
        """Interact with cases existing in the database."""
        # These commands can be run with `scout export cases`
        get_cases_command = ["export", "cases", "--json"]
        if case_id:
            get_cases_command.extend(["--case-id", case_id])

        elif status:
            get_cases_command.extend(["--status", status])

        elif finished:
            get_cases_command.append("--finished")

        if reruns:
            LOG.info("Fetching cases that are reruns")
            get_cases_command.append("--reruns")

        if days_ago:
            get_cases_command.extend(["--within-days", str(days_ago)])

        try:
            self.process.run_command(get_cases_command)
            if not self.process.stdout:
                return []
        except CalledProcessError:
            LOG.info("Could not find cases")
            return []

        cases = []
        for case_export in json.loads(self.process.stdout):
            LOG.info("Validating case %s", case_export.get("_id"))
            case_obj = ScoutExportCase(**case_export)
            cases.append(case_obj)
        return cases

    def get_solved_cases(self, days_ago: int) -> List[ScoutExportCase]:
        """
        Get cases solved within chosen timespan

        Args:
            days_ago (int): Maximum days ago a case has been solved

        Return:
            cases (list): list of cases
        """
        return self.get_cases(status="solved", days_ago=days_ago)

    def upload_delivery_report(self, report_path: str, case_id: str, update: bool = False) -> None:
        """Load a delivery report into a case in the database

        If the report already exists the function will exit.
        If the user want to load a report that is already in the database
        'update' has to be 'True'

        Args:
            report_path (string):       Path to delivery report
            case_id     (string):       Case identifier
            update      (bool):         If an existing report should be replaced

        Returns:
            updated_case(dict)

        """
        # This command can be run with `scout load delivery-report <CASE-ID> <REPORT-PATH>`
        upload_delivery_report_command = ["load", "delivery-report", case_id, report_path]
        if update:
            upload_delivery_report_command.append("--update")

        try:
            LOG.info("Uploading delivery report %s to case %s", report_path, case_id)
            self.process.run_command(upload_delivery_report_command)
        except CalledProcessError:
            LOG.warning("Something went wrong when uploading delivery report")



"""
These are wrapper interfaces the link outside apps and/or the main package
together.

- These tools are allowed to handle cross-talk between services.
- These tools are NOT allowed to do calculations on the own that invoce some
  logic that belongs within another service.
"""



"""Test store base module: MIP get files and build bundle"""
import mock
import pytest

from snapshottest import Snapshot

from cg.constants import MIP_RNA_TAGS, MIP_DNA_TAGS, Pipeline
from cg.meta.store.base import get_tags
import cg.meta.store.base as store_base

from cg.exc import AnalysisDuplicationError, PipelineUnknownError, BundleAlreadyAddedError


@mock.patch("cg.store.Store")
@mock.patch("cg.apps.hk.HousekeeperAPI")
@mock.patch("cg.meta.store.mip.add_mip_analysis")
def test_gather_files_and_bundle_in_hk_bundle_already_added(
    mock_add_mip_analysis, mock_housekeeper, mock_cg_store, config_stream, bundle_data
):
    """
    tests the function gather_files_and_bundle_in_housekeeper
    """
    # GIVEN a MIP RNA analysis config file
    mip_rna_config = config_stream["rna_config_store"]

    # WHEN the bundle has already been added to Housekeeper
    mock_add_mip_analysis.return_value = bundle_data
    mock_housekeeper.add_bundle.return_value = None

    # THEN the BundleAlreadyAddedError exception should be raised
    with pytest.raises(BundleAlreadyAddedError) as exc_info:
        store_base.gather_files_and_bundle_in_housekeeper(
            mip_rna_config, mock_housekeeper, mock_cg_store, workflow=Pipeline.MIP_DNA
        )

    assert exc_info.value.message == "bundle already added"


@mock.patch("cg.store.Store")
@mock.patch("housekeeper.store.models")
@mock.patch("cg.apps.hk.HousekeeperAPI")
@mock.patch("cg.meta.store.base.reset_case_action")
@mock.patch("cg.meta.store.base.add_new_analysis")
@mock.patch("cg.meta.store.mip.add_mip_analysis")
def test_gather_files_and_bundle_in_hk_bundle_new_analysis(
    mock_add_mip_analysis,
    mock_add_new_analysis,
    mock_reset_case_action,
    mock_housekeeper_api,
    mock_housekeeper_store,
    mock_cg_store,
    config_stream,
    bundle_data,
):
    """
    tests the function gather_files_and_bundle_in_housekeeper
    """
    # GIVEN a MIP RNA analysis config file
    mip_rna_config = config_stream["rna_config_store"]

    # WHEN adding a new bundle to Housekeeper
    mock_add_mip_analysis.return_value = bundle_data
    mock_bundle = mock_housekeeper_store.Bundle.return_value
    mock_version = mock_housekeeper_store.Version.return_value
    mock_housekeeper_api.add_bundle.return_value = (mock_bundle, mock_version)
    mock_case = mock_cg_store.Family.return_value
    mock_reset_case_action(mock_case)
    mock_add_new_analysis.return_value = mock_cg_store.Analysis.return_value

    store_base.gather_files_and_bundle_in_housekeeper(
        mip_rna_config, mock_housekeeper_api, mock_cg_store, workflow=Pipeline.MIP_DNA
    )

    # THEN the bundle and version should be added to Housekeeper
    mock_housekeeper_api.include.assert_called()
    mock_housekeeper_api.include.assert_called_with(mock_version)
    mock_housekeeper_api.add_commit.assert_called()
    mock_housekeeper_api.add_commit.assert_called_with(mock_bundle, mock_version)


@mock.patch("cg.store.Store")
@mock.patch("housekeeper.store.models")
def test_add_new_analysis_pipeline_exception(mock_housekeeper_store, mock_status):
    """
    test catching the exception when no pipeline is found when creating and adding a new
    analysis to status-db
    """
    # GIVEN a case for which the bundle and version is added to Housekeeper, but there is no
    # pipeline specified on sample level
    mock_bundle = mock_housekeeper_store.Bundle.return_value
    mock_version = mock_housekeeper_store.Version.return_value
    mock_case = mock_status.Family.return_value
    mock_case.data_analysis = None

    # WHEN creating and adding an analysis object for that case to status-db
    with pytest.raises(PipelineUnknownError) as exc_info:
        store_base.add_new_analysis(
            mock_bundle,
            mock_case,
            mock_status,
            mock_version,
        )

    # THEN a PipelineUnknownError exception should be raised
    assert exc_info.value.message == f"No pipeline specified in {mock_case}"


@mock.patch("cg.store.Store")
@mock.patch("housekeeper.store.models")
def test_add_new_analysis_duplicate_analysis_exception(mock_housekeeper_store, mock_status):
    """
    test catching the exception when no pipeline is found
    """
    # GIVEN a case for which the bundle and version is added to Housekeeper, but the case already
    # has an analysis object stored in status-db
    mock_bundle = mock_housekeeper_store.Bundle.return_value
    mock_version = mock_housekeeper_store.Version.return_value
    mock_case = mock_status.Family.return_value
    mock_case.data_analysis = Pipeline.MIP_DNA
    mock_status.analysis = mock_status.analysis.return_value

    # WHEN creating and adding an analysis object for that case to status-db
    with pytest.raises(AnalysisDuplicationError) as exc_info:
        store_base.add_new_analysis(mock_bundle, mock_case, mock_status, mock_version)

    # THEN an AnalysisDuplicationEror should be raised
    assert (
        exc_info.value.message
        == f"Analysis object already exists for {mock_case.internal_id} {mock_version.created_at}"
    )


@mock.patch("cg.store.Store")
@mock.patch("housekeeper.store.models")
def test_add_new_analysis(mock_housekeeper_store, mock_status):
    """
    test adding a new analysis to cg store
    """
    # GIVEN a case for which the bundle and version is added to Housekeeper
    mock_bundle = mock_housekeeper_store.Bundle.return_value
    mock_version = mock_housekeeper_store.Version.return_value
    mock_case = mock_status.Family.return_value
    mock_status.analysis.return_value = None
    mock_case.data_analysis = Pipeline.MIP_DNA

    # WHEN creating and adding an analysis object for that case in status-db
    new_analysis = store_base.add_new_analysis(
        mock_bundle,
        mock_case,
        mock_status,
        mock_version,
    )

    # THEN an analysis object for that case should created and returned
    assert new_analysis.family == mock_case


@mock.patch("cg.meta.store.base.deliverables_files")
@mock.patch("cg.meta.store.base._determine_missing_tags")
def test_build_rna_bundle(
    mock_missing,
    mock_deliverables_files,
    snapshot: Snapshot,
    config_data_rna: dict,
    sampleinfo_data: dict,
    rna_deliverables_raw: dict,
):
    """
    tests the function_build bundle against a snapshot
    """
    # GIVEN the MIP RNA analysis config data, the sampleinfo data and the deliverables file

    # WHEN building the bundle
    mock_missing.return_value = False, []
    mock_deliverables_files.return_value = {"path": "mock_path"}
    mip_rna_bundle = store_base.build_bundle(config_data_rna, sampleinfo_data, rna_deliverables_raw)

    # THEN the result should contain the data to be stored in Housekeeper
    snapshot.assert_match(mip_rna_bundle)


@mock.patch("cg.meta.store.base.deliverables_files")
@mock.patch("cg.meta.store.base._determine_missing_tags")
def test_build_dna_bundle(
    mock_missing,
    mock_deliverables_files,
    snapshot: Snapshot,
    config_data_dna: dict,
    sampleinfo_data: dict,
    dna_deliverables_raw: dict,
):
    """
    tests the function_build bundle against a snapshot
    """
    # GIVEN the MIP DNA analysis config data, the sampleinfo data and the deliverables file

    # WHEN building the bundle
    mock_missing.return_value = False, []
    mock_deliverables_files.return_value = {"path": "mock_path"}
    mip_dna_bundle = store_base.build_bundle(config_data_dna, sampleinfo_data, dna_deliverables_raw)

    # THEN the result should contain the data to be stored in Housekeeper
    snapshot.assert_match(mip_dna_bundle)


@mock.patch("cg.meta.store.base._determine_missing_tags")
def test_parse_files_rna(mock_missing, snapshot: Snapshot, rna_deliverables_raw: dict):
    """
    tests the function parse_files against a snapshot
    """
    # GIVEN the a MIP RNA analysis deliverables file
    mock_missing.return_value = False, []
    pipeline_tag = ["mip-rna"]
    analysis_type_tags = MIP_RNA_TAGS

    # WHEN getting the files used to build the bundle
    mip_rna_files = store_base.parse_files(rna_deliverables_raw, pipeline_tag, analysis_type_tags)

    # THEN the result should contain the data to be stored in Housekeeper
    snapshot.assert_match(mip_rna_files)


@mock.patch("cg.meta.store.base._determine_missing_tags")
def test_parse_files_dna(mock_missing, snapshot: Snapshot, dna_deliverables_raw: dict):
    """
    tests the function parse_files against a snapshot
    """
    # GIVEN the a MIP DNA analysis deliverables file
    mock_missing.return_value = False, []
    pipeline_tags = ["mip-dna"]
    analysis_type_tags = MIP_DNA_TAGS

    # WHEN getting the files used to build the bundle
    mip_dna_files = store_base.parse_files(dna_deliverables_raw, pipeline_tags, analysis_type_tags)

    # THEN the result should contain the data to be stored in Housekeeper
    snapshot.assert_match(mip_dna_files)


@mock.patch("cg.meta.store.base._determine_missing_tags")
def test_get_tags(mock_missing, dna_deliverables_raw: dict):
    """
    Tests get converted housekeeper tags
    """
    # GIVEN a MIP DNA analysis deliverables file, pipeline tags and corresponding analysis_tags
    mock_missing.return_value = False, []
    pipeline_tags = ["mip-dna"]
    analysis_type_tags = MIP_DNA_TAGS

    # WHEN getting the tags used to build the bundle
    for file_ in dna_deliverables_raw["files"]:
        deliverables_tag_map = (
            (file_["step"],) if file_["tag"] is None else (file_["step"], file_["tag"])
        )
        hk_tags = get_tags(file_, pipeline_tags, analysis_type_tags, deliverables_tag_map)
        # Then only analysis tags, pipeline tag and sample_id should be returned
        if file_["step"] == "samtools_subsample_mt":
            assert hk_tags == ["bam-mt", "mip-dna", "sample_id"]
        if file_["step"] == "gatk_baserecalibration":
            assert hk_tags == ["cram", "mip-dna", "sample_id"]


@mock.patch("cg.meta.store.base._determine_missing_tags")
def test_get_tags_for_index_file(mock_missing, dna_deliverables_raw: dict):
    """
    Tests get converted housekeeper tags for index file
    """
    # GIVEN a MIP DNA analysis deliverables file, pipeline tags and corresponding analysis_tags
    mock_missing.return_value = False, []
    pipeline_tags = ["mip-dna"]
    analysis_type_tags = MIP_DNA_TAGS

    # WHEN getting the tags used to build the bundle
    for file_ in dna_deliverables_raw["files"]:
        if file_["path_index"]:
            deliverables_tag_map = (
                (file_["step"],) if file_["tag"] is None else (file_["step"], file_["tag"])
            )
            hk_tags = get_tags(
                file_, pipeline_tags, analysis_type_tags, deliverables_tag_map, is_index=True
            )
            # Then only analysis tags, pipeline tag and sample_id should be returned
            if file_["step"] == "samtools_subsample_mt":
                assert hk_tags == ["bam-mt-index", "mip-dna", "sample_id"]
            if file_["step"] == "gatk_baserecalibration":
                assert hk_tags == ["cram-index", "mip-dna", "sample_id"]



"""Fixtures for the meta tests"""

from datetime import datetime

import pytest
from _pytest import tmpdir

from cg.store import Store
from cg.apps.balsamic.fastq import FastqHandler as BalsamicFastqHandler
from cg.apps.microsalt.fastq import FastqHandler as MicrosaltFastqHandler
from cg.apps.hk import HousekeeperAPI
from cg.apps.scout.scoutapi import ScoutAPI
from cg.meta.workflow.mip import MipAnalysisAPI

from tests.store_helpers import StoreHelpers


@pytest.yield_fixture(scope="function", name="analysis_store")
def fixture_analysis_store(base_store: Store, analysis_family: dict) -> Store:
    """Setup a store instance for testing analysis API."""
    customer = base_store.customer("cust000")
    family = base_store.Family(
        data_analysis=analysis_family["data_analysis"],
        name=analysis_family["name"],
        panels=analysis_family["panels"],
        internal_id=analysis_family["internal_id"],
        priority="standard",
    )
    family.customer = customer
    base_store.add(family)
    application_version = base_store.application("WGTPCFC030").versions[0]
    for sample_data in analysis_family["samples"]:
        sample = base_store.add_sample(
            name=sample_data["name"],
            sex=sample_data["sex"],
            internal_id=sample_data["internal_id"],
            ticket=sample_data["ticket_number"],
            reads=sample_data["reads"],
            capture_kit=sample_data["capture_kit"],
        )
        sample.family = family
        sample.application_version = application_version
        sample.customer = customer
        base_store.add(sample)
    base_store.commit()
    for sample_data in analysis_family["samples"]:
        sample_obj = base_store.sample(sample_data["internal_id"])
        link = base_store.relate_sample(
            family=family,
            sample=sample_obj,
            status=sample_data["status"],
            father=base_store.sample(sample_data["father"]) if sample_data.get("father") else None,
            mother=base_store.sample(sample_data["mother"]) if sample_data.get("mother") else None,
        )
        base_store.add(link)
    base_store.commit()
    yield base_store


class MockLims:
    """Mock lims fixture"""

    lims = None

    def __init__(self):
        self.lims = self


class MockPath:
    def __init__(self, path):
        self.yaml = None

    def __call__(self, *args, **kwargs):
        return self

    def open(self):
        return dict()

    def joinpath(self, path):
        return path


class MockLogger:
    warnings = []

    def warning(self, text, *interpolations):

        self.warnings.append(text % interpolations)

    def get_warnings(self) -> list:
        return self.warnings


class MockTB:
    _get_trending_raises_keyerror = False

    def __init__(self):
        """Needed to initialise mock variables"""
        self._make_config_was_called = False

    def get_trending(self, mip_config_raw: dict, qcmetrics_raw: dict, sampleinfo_raw: dict) -> dict:
        if self._get_trending_raises_keyerror:
            raise KeyError("mockmessage")

        # Returns: dict: parsed data
        ### Define output dict
        outdata = {
            "analysis_sex": {"ADM1": "female", "ADM2": "female", "ADM3": "female"},
            "family": "yellowhog",
            "duplicates": {"ADM1": 13.525, "ADM2": 12.525, "ADM3": 14.525},
            "mapped_reads": {"ADM1": 98.8, "ADM2": 99.8, "ADM3": 97.8},
            "mip_version": "v4.0.20",
            "sample_ids": ["2018-20203", "2018-20204"],
            "genome_build": "37",
            "rank_model_version": "1.18",
        }

        return outdata

    def get_sampleinfo(self, analysis_obj):
        return ""

    def make_config(self, data, pipeline):
        """Mock the make_config"""
        self._make_config_was_called = True
        del pipeline
        return data


class MockBalsamicFastq(BalsamicFastqHandler):
    """Mock FastqHandler for analysis_api"""

    def __init__(self):
        super().__init__(config={"balsamic": {"root": tmpdir}})


class MockMicrosaltFastq(MicrosaltFastqHandler):
    """Mock FastqHandler for analysis_api"""

    def __init__(self):
        super().__init__(config={"microsalt": {"root": tmpdir}})


@pytest.yield_fixture(scope="function", name="mip_hk_store")
def fixture_mip_hk_store(
    helpers: StoreHelpers,
    real_housekeeper_api: HousekeeperAPI,
    timestamp: datetime,
    case_id: str,
) -> HousekeeperAPI:
    deliver_hk_bundle_data = {
        "name": case_id,
        "created": timestamp,
        "expires": timestamp,
        "files": [
            {
                "path": "tests/fixtures/apps/mip/dna/store/case_config.yaml",
                "archive": False,
                "tags": ["mip-config"],
            },
            {
                "path": "tests/fixtures/apps/mip/dna/store/case_qc_sample_info.yaml",
                "archive": False,
                "tags": ["sampleinfo"],
            },
            {
                "path": "tests/fixtures/apps/mip/case_qc_metrics.yaml",
                "archive": False,
                "tags": ["qcmetrics"],
            },
            {
                "path": "tests/fixtures/apps/mip/case_file.txt",
                "archive": False,
                "tags": ["case-tag"],
            },
            {
                "path": "tests/fixtures/apps/mip/sample_file.txt",
                "archive": False,
                "tags": ["sample-tag", "ADM1"],
            },
        ],
    }
    helpers.ensure_hk_bundle(real_housekeeper_api, deliver_hk_bundle_data, include=True)

    empty_deliver_hk_bundle_data = {
        "name": "case_missing_data",
        "created": timestamp,
        "expires": timestamp,
        "files": [
            {
                "path": "tests/fixtures/apps/mip/dna/store/empty_case_config.yaml",
                "archive": False,
                "tags": ["mip-config"],
            },
            {
                "path": "tests/fixtures/apps/mip/dna/store/empty_case_qc_sample_info.yaml",
                "archive": False,
                "tags": ["sampleinfo"],
            },
            {
                "path": "tests/fixtures/apps/mip/dna/store/empty_case_qc_metrics.yaml",
                "archive": False,
                "tags": ["qcmetrics"],
            },
            {
                "path": "tests/fixtures/apps/mip/case_file.txt",
                "archive": False,
                "tags": ["case-tag"],
            },
            {
                "path": "tests/fixtures/apps/mip/sample_file.txt",
                "archive": False,
                "tags": ["sample-tag", "ADM1"],
            },
        ],
    }
    helpers.ensure_hk_bundle(real_housekeeper_api, empty_deliver_hk_bundle_data, include=True)

    return real_housekeeper_api


@pytest.yield_fixture(scope="function", name="analysis_api")
def fixture_analysis_api(
    analysis_store: Store, mip_hk_store: HousekeeperAPI, scout_api: ScoutAPI
) -> MipAnalysisAPI:
    """Setup an analysis API."""

    analysis_api = MipAnalysisAPI(
        db=analysis_store,
        hk_api=mip_hk_store,
        scout_api=scout_api,
        tb_api=MockTB(),
        lims_api=MockLims(),
        script="echo",
        pipeline="analyse rd_dna",
        conda_env="S_mip_rd-dna",
        root="/var/empty",
    )
    yield analysis_api



from cg.store import Store


def test_microbial_sample_to_dict(microbial_store: Store, helpers):

    # GIVEN a store with a Microbial sample
    sample_obj = helpers.add_microbial_sample(microbial_store)
    microbial_store.add_commit(sample_obj)
    assert sample_obj

    # WHEN running to dict on that sample
    a_dict = sample_obj.to_dict(links=True)

    # THEN you should get a dictionary with
    assert a_dict["id"]
    assert a_dict["internal_id"]
    assert a_dict["name"]
    assert a_dict["application_version_id"]
    assert a_dict["created_at"]
    assert a_dict["priority"]
    assert a_dict["reads"]
    assert a_dict["comment"]
    assert a_dict["application"]
    assert a_dict["application_version"]



"""CLI support to create config and/or start BALSAMIC """

import logging

import click

from cg.apps.environ import environ_email
from cg.apps.balsamic.api import BalsamicAPI
from cg.apps.balsamic.fastq import FastqHandler
from cg.apps.hk import HousekeeperAPI
from cg.apps.lims import LimsAPI
from cg.apps.tb import TrailblazerAPI
from cg.exc import BalsamicStartError, BundleAlreadyAddedError, LimsDataError
from cg.meta.workflow.balsamic import BalsamicAnalysisAPI
from cg.store import Store
from cg.constants import EXIT_SUCCESS, EXIT_FAIL, Pipeline

LOG = logging.getLogger(__name__)

ARGUMENT_CASE_ID = click.argument("case_id", required=True)
OPTION_DRY = click.option(
    "-d", "--dry-run", "dry", help="Print command to console without executing", is_flag=True
)
OPTION_PANEL_BED = click.option(
    "--panel-bed",
    required=False,
    help="Panel BED is determined based on capture kit \
    used for library prep. Set this option to override the default",
)
OPTION_ANALYSIS_TYPE = click.option(
    "-a",
    "--analysis-type",
    type=click.Choice(["qc", "paired", "single"]),
    help="Setting this option to qc ensures only QC analysis is performed",
)
OPTION_RUN_ANALYSIS = click.option(
    "-r",
    "--run-analysis",
    is_flag=True,
    default=False,
    help="Execute BALSAMIC in non-dry mode",
)
OPTION_PRIORITY = click.option(
    "-p",
    "--priority",
    type=click.Choice(["low", "normal", "high"]),
    help="Job priority in SLURM. Will be set automatically according to priority in ClinicalDB, \
         this option can be used to override server setting",
)


@click.group(invoke_without_command=True)
@click.pass_context
def balsamic(context):
    """Cancer analysis workflow """
    if context.invoked_subcommand is None:
        click.echo(context.get_help())
        return None
    config = context.obj
    context.obj["BalsamicAnalysisAPI"] = BalsamicAnalysisAPI(
        balsamic_api=BalsamicAPI(config),
        store=Store(config["database"]),
        housekeeper_api=HousekeeperAPI(config),
        fastq_handler=FastqHandler(config),
        lims_api=LimsAPI(config),
        trailblazer_api=TrailblazerAPI(config),
    )


@balsamic.command("link")
@ARGUMENT_CASE_ID
@click.pass_context
def link(context, case_id):
    """
    Locates FASTQ files for given CASE_ID.
    The files are renamed, concatenated, and saved in BALSAMIC working directory
    """
    balsamic_analysis_api = context.obj["BalsamicAnalysisAPI"]
    try:
        LOG.info(f"Linking samples in case {case_id}")
        balsamic_analysis_api.link_samples(case_id)
    except BalsamicStartError as e:
        LOG.error(f"Could not link samples: {e.message}")
        raise click.Abort()


@balsamic.command("config-case")
@ARGUMENT_CASE_ID
@OPTION_PANEL_BED
@OPTION_DRY
@click.pass_context
def config_case(context, panel_bed, case_id, dry):
    """Create config file for BALSAMIC analysis for a given CASE_ID"""
    balsamic_analysis_api = context.obj["BalsamicAnalysisAPI"]

    try:
        LOG.info(f"Creating config file for {case_id}.")
        arguments = balsamic_analysis_api.get_verified_config_case_arguments(
            case_id=case_id,
            panel_bed=panel_bed,
        )
        balsamic_analysis_api.balsamic_api.config_case(arguments=arguments, dry=dry)
    except (BalsamicStartError, LimsDataError) as e:
        LOG.error(f"Could not create config: {e.message}")
        raise click.Abort()


@balsamic.command("run")
@ARGUMENT_CASE_ID
@OPTION_DRY
@OPTION_PRIORITY
@OPTION_ANALYSIS_TYPE
@OPTION_RUN_ANALYSIS
@click.pass_context
def run(context, analysis_type, run_analysis, priority, case_id, dry):
    """Run balsamic analysis for given CASE ID"""
    balsamic_analysis_api = context.obj["BalsamicAnalysisAPI"]
    try:
        LOG.info(f"Running analysis for {case_id}")
        if balsamic_analysis_api.trailblazer_api.is_latest_analysis_ongoing(case_id=case_id):
            LOG.warning(f"{case_id} : analysis is still ongoing - skipping")
            return

        arguments = {
            "priority": priority or balsamic_analysis_api.get_priority(case_id),
            "analysis_type": analysis_type,
            "run_analysis": run_analysis,
            "sample_config": balsamic_analysis_api.get_config_path(
                case_id=case_id, check_exists=True
            ),
            "disable_variant_caller": "mutect"
            if balsamic_analysis_api.get_case_application_type(case_id=case_id) == "wes"
            else None,  # Tell Balsamic to disable mutect for WES analyses.
        }
        balsamic_analysis_api.balsamic_api.run_analysis(
            arguments=arguments, run_analysis=run_analysis, dry=dry
        )
        balsamic_analysis_api.trailblazer_api.mark_analyses_deleted(case_id=case_id)
        balsamic_analysis_api.trailblazer_api.add_pending_analysis(
            case_id=case_id,
            email=environ_email(),
            type=balsamic_analysis_api.get_case_application_type(case_id=case_id),
            out_dir=balsamic_analysis_api.get_case_path(case_id),
            config_path=balsamic_analysis_api.get_slurm_job_ids_path(case_id).as_posix(),
            priority=balsamic_analysis_api.get_priority(case_id),
            data_analysis=Pipeline.BALSAMIC,
        )
        balsamic_analysis_api.set_statusdb_action(case_id=case_id, action="running")
    except BalsamicStartError as e:
        LOG.error(f"Could not run analysis: {e.message}")
        raise click.Abort()


@balsamic.command("report-deliver")
@ARGUMENT_CASE_ID
@OPTION_DRY
@OPTION_ANALYSIS_TYPE
@click.pass_context
def report_deliver(context, case_id, analysis_type, dry):
    """Create a housekeeper deliverables file for given CASE ID"""
    balsamic_analysis_api = context.obj["BalsamicAnalysisAPI"]
    try:
        LOG.info(f"Creating delivery report for {case_id}")
        case_object = balsamic_analysis_api.get_case_object(case_id)
        sample_config = balsamic_analysis_api.get_config_path(case_id=case_id, check_exists=True)
        analysis_finish = balsamic_analysis_api.get_analysis_finish_path(case_id, check_exists=True)
        LOG.info(f"Found analysis finish file: {analysis_finish}")
        arguments = {"sample_config": sample_config, "analysis_type": analysis_type}
        balsamic_analysis_api.balsamic_api.report_deliver(arguments=arguments, dry=dry)
    except BalsamicStartError as e:
        LOG.error(f"Could not create report file: {e.message}")
        raise click.Abort()


@balsamic.command("store-housekeeper")
@ARGUMENT_CASE_ID
@click.pass_context
def store_housekeeper(context, case_id):
    """Store a finished analysis in Housekeeper and StatusDB."""
    balsamic_analysis_api = context.obj["BalsamicAnalysisAPI"]
    try:
        LOG.info(f"Storing bundle data in Housekeeper for {case_id}")
        balsamic_analysis_api.upload_bundle_housekeeper(case_id=case_id)
        LOG.info(f"Storing Analysis in ClinicalDB for {case_id}")
        balsamic_analysis_api.upload_analysis_statusdb(case_id=case_id)
    except (BundleAlreadyAddedError, FileExistsError) as e:
        LOG.error(f"Could not store bundle in Housekeeper and StatusDB: {e.message}!")
        balsamic_analysis_api.housekeeper_api.rollback()
        balsamic_analysis_api.store.rollback()
        raise click.Abort()
    except BalsamicStartError as e:
        LOG.error(f"Could not store bundle in Housekeeper and StatusDB: {e.message}!")
        raise click.Abort()


@balsamic.command("start")
@ARGUMENT_CASE_ID
@OPTION_ANALYSIS_TYPE
@OPTION_PRIORITY
@OPTION_DRY
@OPTION_PANEL_BED
@click.pass_context
def start(context, case_id, analysis_type, panel_bed, priority, dry):
    """Start full workflow for CASE ID"""
    LOG.info(f"Starting analysis for {case_id}")
    context.invoke(link, case_id=case_id)
    context.invoke(config_case, case_id=case_id, panel_bed=panel_bed, dry=dry)
    context.invoke(
        run,
        case_id=case_id,
        analysis_type=analysis_type,
        priority=priority,
        run_analysis="--run-analysis",
        dry=dry,
    )


@balsamic.command("store")
@ARGUMENT_CASE_ID
@OPTION_DRY
@OPTION_ANALYSIS_TYPE
@click.pass_context
def store(context, case_id, analysis_type, dry):
    """Generate Housekeeper report for CASE ID and store in Housekeeper"""
    LOG.info(f"Storing analysis for {case_id}")
    context.invoke(report_deliver, case_id=case_id, analysis_type=analysis_type, dry=dry)
    context.invoke(store_housekeeper, case_id=case_id)


@balsamic.command("start-available")
@OPTION_DRY
@click.pass_context
def start_available(context, dry):
    """Start full workflow for all available BALSAMIC cases"""
    balsamic_analysis_api = context.obj["BalsamicAnalysisAPI"]
    exit_code = EXIT_SUCCESS
    for case_id in balsamic_analysis_api.get_cases_to_analyze():
        try:
            context.invoke(start, case_id=case_id, dry=dry)
        except click.Abort:
            exit_code = EXIT_FAIL
        except Exception as e:
            LOG.error(f"Unspecified error occurred - {e.__class__.__name__}")
            exit_code = EXIT_FAIL
    if exit_code:
        raise click.Abort()


@balsamic.command("store-available")
@OPTION_DRY
@click.pass_context
def store_available(context, dry):
    """Store bundle data for all available Balsamic cases"""
    balsamic_analysis_api = context.obj["BalsamicAnalysisAPI"]
    exit_code = EXIT_SUCCESS
    for case_id in balsamic_analysis_api.get_cases_to_store():
        try:
            context.invoke(store, case_id=case_id, dry=dry)
        except click.Abort:
            exit_code = EXIT_FAIL
        except Exception as e:
            LOG.error(f"Unspecified error occurred - {e.__class__.__name__}")
            exit_code = EXIT_FAIL
    if exit_code:
        raise click.Abort()



"""Tags for storing analyses in Housekeeper"""

HK_TAGS = {
    "wes": ["mip-dna", "wes"],
    "wgs": ["mip-dna", "wgs"],
    "wts": ["mip-rna"],
    "microbial": ["microsalt"],
}
HK_FASTQ_TAGS = ["fastq"]

# used to convert MIP tags derived from the deliverables to MIP standard tags and to check for
# presence of mandatory files. Keys = tags found in deliverables, values = MIP standard tags and
# mandatory flag
MIP_DNA_TAGS = {
    ("chanjo_sexcheck",): {"tags": ["chanjo", "sex-check"], "is_mandatory": False},
    ("chromograph_upd", "sites"): {"tags": ["chromograph", "upd", "sites"], "is_mandatory": False},
    ("chromograph_upd", "regions"): {
        "tags": ["chromograph", "upd", "regions"],
        "is_mandatory": False,
    },
    ("chromograph_cov", "tcov"): {"tags": ["chromograph", "tcov"], "is_mandatory": False},
    ("endvariantannotationblock", "clinical"): {
        "tags": ["vcf-snv-clinical"],
        "index_tags": ["vcf-snv-clinical-index"],
        "is_mandatory": True,
    },
    ("endvariantannotationblock", "research"): {
        "tags": ["vcf-snv-research"],
        "index_tags": ["vcf-snv-research-index"],
        "is_mandatory": True,
    },
    ("expansionhunter", "sv_str"): {
        "tags": ["vcf-str"],
        "index_tags": ["vcf-str-index"],
        "is_mandatory": False,
    },
    ("gatk_baserecalibration",): {
        "tags": ["cram"],
        "index_tags": ["cram-index"],
        "is_mandatory": False,
    },
    ("gatk_combinevariantcallsets",): {
        "tags": ["snv-gbcf", "snv-bcf"],
        "index_tags": ["gbcf-index"],
        "is_mandatory": True,
    },
    ("mip_analyse", "config"): {"tags": ["mip-analyse", "config"], "is_mandatory": True},
    ("mip_analyse", "config_analysis"): {"tags": ["mip-config"], "is_mandatory": True},
    ("mip_analyse", "log"): {"tags": ["mip-log"], "is_mandatory": True},
    ("mip_analyse", "pedigree"): {"tags": ["pedigree-yaml"], "is_mandatory": True},
    ("mip_analyse", "pedigree_fam"): {"tags": ["pedigree"], "is_mandatory": True},
    ("mip_analyse", "references_info"): {
        "tags": ["mip-analyse", "reference-info"],
        "is_mandatory": True,
    },
    ("mip_analyse", "sample_info"): {"tags": ["sampleinfo"], "is_mandatory": True},
    ("multiqc_ar", "html"): {"tags": ["multiqc-html"], "is_mandatory": True},
    ("multiqc_ar", "json"): {"tags": ["multiqc-json"], "is_mandatory": True},
    ("peddy_ar", "ped_check"): {"tags": ["peddy", "ped-check"], "is_mandatory": True},
    ("peddy_ar", "peddy"): {"tags": ["peddy", "ped"], "is_mandatory": True},
    ("peddy_ar", "sex_check"): {"tags": ["peddy", "sex-check"], "is_mandatory": True},
    ("qccollect_ar",): {"tags": ["qcmetrics"], "is_mandatory": True},
    ("rhocall_viz",): {"tags": ["rhocall-viz"], "is_mandatory": False},
    ("sambamba_depth", "coverage"): {"tags": ["coverage", "sambamba-depth"], "is_mandatory": True},
    ("samtools_subsample_mt",): {
        "tags": ["bam-mt"],
        "index_tags": ["bam-mt-index"],
        "is_mandatory": False,
    },
    ("smncopynumbercaller",): {
        "tags": ["smn-calling", "smncopynumbercaller"],
        "is_mandatory": False,
    },
    ("star_caller",): {
        "tags": ["cyrius", "star-caller"],
        "is_mandatory": False,
    },
    ("sv_combinevariantcallsets",): {
        "tags": ["sv-bcf"],
        "index_tags": ["sv-bcf-index"],
        "is_mandatory": True,
    },
    ("sv_reformat", "clinical"): {
        "tags": ["vcf-sv-clinical"],
        "index_tags": ["vcf-sv-clinical-index"],
        "is_mandatory": False,
    },
    ("sv_reformat", "research"): {
        "tags": ["vcf-sv-research"],
        "index_tags": ["vcf-sv-research-index"],
        "is_mandatory": False,
    },
    ("telomerecat_ar",): {
        "tags": ["telomere-calling", "telomerecat"],
        "is_mandatory": False,
    },
    ("tiddit_coverage",): {
        "tags": ["tiddit-coverage", "bigwig"],
        "is_mandatory": False,
    },
    ("upd_ar", "regions"): {"tags": ["upd", "regions"], "is_mandatory": False},
    ("upd_ar", "sites"): {"tags": ["upd", "sites"], "is_mandatory": False},
    ("version_collect_ar",): {"tags": ["exe-ver"], "is_mandatory": True},
    ("vcf2cytosure_ar",): {"tags": ["vcf2cytosure"], "is_mandatory": False},
}


MIP_RNA_TAGS = {
    ("arriba_ar", "arriba_ar"): {"tags": ["arriba-ar"], "is_mandatory": True},
    ("arriba_ar", "arriba_report"): {"tags": ["arriba-ar", "arriba-report"], "is_mandatory": True},
    ("bcftools_merge",): {"tags": ["bcftools-merge"], "is_mandatory": True},
    ("blobfish",): {"tags": ["blobfish"], "is_mandatory": False},
    ("bootstrapann",): {"tags": ["bootstrapann"], "is_mandatory": True},
    ("gatk_asereadcounter",): {"tags": ["gatk-asereadcounter"], "is_mandatory": True},
    ("gffcompare_ar",): {"tags": ["gffcompare-ar"], "is_mandatory": True},
    ("markduplicates",): {"tags": ["cram"], "index_tags": ["cram-index"], "is_mandatory": True},
    ("mip_analyse", "config"): {"tags": ["mip-analyse", "config"], "is_mandatory": True},
    ("mip_analyse", "config_analysis"): {
        "tags": ["mip-analyse", "config-analysis"],
        "is_mandatory": True,
    },
    ("mip_analyse", "log"): {"tags": ["mip-analyse", "log"], "is_mandatory": True},
    ("mip_analyse", "pedigree"): {"tags": ["mip-analyse", "pedigree"], "is_mandatory": True},
    ("mip_analyse", "pedigree_fam"): {
        "tags": ["mip-analyse", "pedigree-fam"],
        "is_mandatory": True,
    },
    ("mip_analyse", "references_info"): {
        "tags": ["mip-analyse", "reference-info"],
        "is_mandatory": True,
    },
    ("mip_analyse", "sample_info"): {"tags": ["mip-analyse", "sample-info"], "is_mandatory": True},
    ("multiqc_ar", "html"): {"tags": ["multiqc-html"], "is_mandatory": True},
    ("multiqc_ar", "json"): {"tags": ["multiqc-json"], "is_mandatory": True},
    ("salmon_quant",): {"tags": ["salmon-quant"], "is_mandatory": True},
    ("star_fusion",): {"tags": ["star-fusion"], "is_mandatory": True},
    ("stringtie_ar",): {"tags": ["stringtie-ar"], "is_mandatory": True},
    ("varianteffectpredictor",): {"tags": ["varianteffectpredictor"], "is_mandatory": True},
    ("version_collect_ar",): {"tags": ["version-collect-ar"], "is_mandatory": True},
}

MICROSALT_TAGS = {
    ("alignment", "reference-alignment"): {
        "tags": ["alignment", "reference-alignment"],
        "is_mandatory": True,
    },
    ("alignment", "reference-alignment-deduplicated"): {
        "tags": ["alignment", "reference-alignment-deduplicated"],
        "index_tags": ["reference-alignment-deduplicate-index"],
        "is_mandatory": True,
    },
    ("alignment", "reference-alignment-sorted"): {
        "tags": ["alignment", "reference-alignment-sorted"],
        "is_mandatory": True,
    },
    ("analysis", "logfile"): {
        "tags": ["analysis", "log"],
        "is_mandatory": True,
    },
    ("analysis", "runtime-settings"): {
        "tags": ["analysis", "runtime-settings"],
        "is_mandatory": True,
    },
    ("analysis", "sampleinfo"): {
        "tags": ["analysis", "sampleinfo"],
        "is_mandatory": True,
    },
    ("assembly", "assembly"): {
        "tags": ["assembly"],
        "is_mandatory": True,
    },
    ("assembly", "quast-results"): {
        "tags": ["assembly", "quast-results"],
        "is_mandatory": True,
    },
    ("concatination", "trimmed-forward-reads"): {
        "tags": ["concatination", "trimmed-forward-reads"],
        "is_mandatory": True,
    },
    ("concatination", "trimmed-reverse-reads"): {
        "tags": ["concatination", "trimmed-reverse-reads"],
        "is_mandatory": True,
    },
    ("result_aggregation", "microsalt-json"): {
        "tags": ["result-aggregation-json"],
        "is_mandatory": True,
    },
    ("result_aggregation", "microsalt-qc"): {
        "tags": ["result-aggregration-qc"],
        "is_mandatory": True,
    },
    ("result_aggregation", "microsalt-type"): {
        "tags": ["result-aggregation-type"],
        "is_mandatory": True,
    },
    ("insertsize_calc", "picard-insertsize"): {
        "tags": ["insertsize-calc", "picard-insertsize"],
        "is_mandatory": True,
    },
}



"""Class that fetches and calculates data from status database """
from cg.store import models


class ReportHelper:
    """Class that fetches and calculates data from status database """

    @staticmethod
    def get_report_version(analysis: models.Analysis) -> int:
        """Returns the version for the given analysis where the primary analysis is 1 and
        subsequent reruns increase the version by for each"""

        version = None

        if analysis:
            version = len(analysis.family.analyses) - analysis.family.analyses.index(analysis)

        return version

    @staticmethod
    def get_previous_report_version(analysis: models.Analysis) -> int:
        """Returns the version for the analysis before the given analysis"""

        analysis_version = ReportHelper.get_report_version(analysis)
        previous_version = None

        if analysis_version and analysis_version > 1:
            previous_version = analysis_version - 1

        return previous_version



from cg.store import Store


def test_get_family_by_name(invoke_cli, disk_store: Store):
    # GIVEN A database with a customer in it
    customer_id = "customer-test"
    customer_group = disk_store.add_customer_group("dummy_group", "dummy group")
    customer = disk_store.add_customer(
        internal_id=customer_id,
        name="Test Customer",
        scout_access=False,
        customer_group=customer_group,
        invoice_address="Street nr, 12345 Uppsala",
        invoice_reference="ABCDEF",
    )
    disk_store.add_commit(customer)

    # WHEN trying to get a non-existing family by name
    db_uri = disk_store.uri
    result = invoke_cli(
        [
            "--database",
            db_uri,
            "get",
            "family",
            "-c",
            customer_id,
            "-n",
            "dummy-family-name",
        ]
    )

    # THEN it should not crash
    assert result.exit_code == 0



from copy import deepcopy
import logging
from pathlib import Path

from marshmallow import Schema, fields, validate
import ruamel.yaml

from cg.exc import PedigreeConfigError
from cg.constants import DEFAULT_CAPTURE_KIT, NO_PARENT, Pipeline

LOG = logging.getLogger(__name__)


class SampleSchema(Schema):
    sample_id = fields.Str(required=True)
    sample_display_name = fields.Str()
    analysis_type = fields.Str(
        required=True,
        validate=validate.OneOf(
            choices=[
                "tga",
                "wes",
                "wgs",
                "wts",
            ]
        ),
    )
    father = fields.Str(default=NO_PARENT)
    mother = fields.Str(default=NO_PARENT)
    phenotype = fields.Str(
        required=True,
        validate=validate.OneOf(choices=["affected", "unaffected", "unknown"]),
    )
    sex = fields.Str(required=True, validate=validate.OneOf(choices=["female", "male", "unknown"]))
    expected_coverage = fields.Float()
    capture_kit = fields.Str(default=DEFAULT_CAPTURE_KIT)


class ConfigSchema(Schema):
    case = fields.Str(required=True)
    default_gene_panels = fields.List(fields.Str(), required=True)
    samples = fields.List(fields.Nested(SampleSchema), required=True)


class ConfigHandler:
    def make_pedigree_config(self, data: dict, pipeline: Pipeline = None) -> dict:
        """Make a MIP pedigree config"""
        self.validate_config(data=data, pipeline=pipeline)
        config_data = self.parse_pedigree_config(data)
        return config_data

    @staticmethod
    def validate_config(data: dict, pipeline: Pipeline = None) -> dict:
        """Validate MIP pedigree config format"""
        errors = ConfigSchema().validate(data)
        fatal_error = False
        for field, messages in errors.items():
            if isinstance(messages, dict):
                for sample_index, sample_errors in messages.items():
                    try:
                        sample_id = data["samples"][sample_index]["sample_id"]
                    except KeyError:
                        raise PedigreeConfigError("missing sample id")
                    for sample_key, sub_messages in sample_errors.items():
                        if sub_messages != ["Unknown field."]:
                            fatal_error = True
                        LOG.error(f"{sample_id} -> {sample_key}: {', '.join(sub_messages)}")
            else:
                fatal_error = True
                LOG.error(f"{field}: {', '.join(messages)}")
        if fatal_error:
            raise PedigreeConfigError("invalid config input", errors=errors)
        return errors

    @staticmethod
    def parse_pedigree_config(data: dict) -> dict:
        """Parse the pedigree config data"""
        data_copy = deepcopy(data)
        # handle single sample cases with 'unknown' phenotype
        if len(data_copy["samples"]) == 1 and data_copy["samples"][0]["phenotype"] == "unknown":
            LOG.info("setting 'unknown' phenotype to 'unaffected'")
            data_copy["samples"][0]["phenotype"] = "unaffected"
        for sample_data in data_copy["samples"]:
            sample_data["mother"] = sample_data.get("mother") or NO_PARENT
            sample_data["father"] = sample_data.get("father") or NO_PARENT
            if sample_data["analysis_type"] == "wgs" and sample_data.get("capture_kit") is None:
                sample_data["capture_kit"] = DEFAULT_CAPTURE_KIT
        return data_copy

    @staticmethod
    def write_pedigree_config(data: dict, out_dir: Path, pedigree_config_path: Path) -> Path:
        """Write the pedigree config to the the case dir"""
        out_dir.mkdir(parents=True, exist_ok=True)
        dump = ruamel.yaml.round_trip_dump(data, indent=4, block_seq_indent=2)
        pedigree_config_path.write_text(dump)
        return pedigree_config_path



"""Models to help parsing relevant data"""

import marshmallow as ma

from cg.utils.date import SIMPLE_DATE_FORMAT


class CrunchyFileSchema(ma.Schema):
    """Schema with information from spring metadata info"""

    path = ma.fields.Str(required=True, error_messages={"required": "Path is required."})
    file = ma.fields.Str(
        required=True,
        validate=ma.validate.OneOf(["first_read", "second_read", "spring"]),
        error_messages={"required": "File needs to be specified."},
    )
    checksum = ma.fields.Str()
    algorithm = ma.fields.Str()
    updated = ma.fields.DateTime(SIMPLE_DATE_FORMAT)



"""Test FastqHandler"""
import os
from pathlib import Path

from cg.apps.microsalt.fastq import FastqHandler


def test_fastq_link_file_count(tmpdir, cg_config, link_ticket, link_sample, simple_files_data):
    """Test method to test that the right number of files are created by linking"""

    # given some fastq-files belonging to family and sample
    link_files = simple_files_data
    assert len(tmpdir.listdir()) == len(link_files)

    # The fastq files should be linked to /.../fastq/<project>/<sample>/*.fastq.gz.
    link_dir = Path(f"{tmpdir}/fastq/{link_ticket}/{link_sample}")
    assert not os.path.exists(link_dir)

    # when calling the method to link
    FastqHandler(cg_config).link(ticket=link_ticket, sample=link_sample, files=link_files)

    # then the linking should have created on directory for the linked files
    assert os.path.exists(link_dir)
    assert len(tmpdir.listdir()) == len(link_files) + 1



""" Test the CLI for run mip-rna """
import logging

from cg.cli.workflow.mip_rna.base import run

CASE_ID = "yellowhog"
CONDA_ENV = "S_mip_rd-rna"
EMAIL = "james.holden@scilifelab.se"


def test_cg_dry_run(cli_runner, tb_api, mip_context, caplog):
    """Test print the MIP command to console"""
    # GIVEN a cli function

    # WHEN we run a case in dry run mode
    caplog.set_level(logging.INFO)
    result = cli_runner.invoke(run, ["--dry-run", "--email", EMAIL, CASE_ID], obj=mip_context)

    # THEN the command should be printed
    with caplog.at_level(logging.INFO):
        assert result.exit_code == 0
        assert (
            "analyse rd_rna yellowhog --config config.yaml "
            "--email james.holden@scilifelab.se" in caplog.text
        )



"""Tests for the compress fastq cli"""

import logging

from cg.cli.compress.fastq import (
    decompress_sample,
    decompress_flowcell,
    decompress_ticket,
    decompress_case,
)


def test_decompress_spring_cli_no_family(compress_context, case_id, cli_runner, caplog):
    """Test to run the decompress command when no families are found"""
    caplog.set_level(logging.DEBUG)
    # GIVEN a context without families

    # WHEN running the compress command
    res = cli_runner.invoke(decompress_case, ["nonexisting_case"], obj=compress_context)

    # THEN assert the program exits since no cases were found
    assert res.exit_code == 0
    # THEN assert it was communicated that no families where found
    assert "Could not find case" in caplog.text


def test_decompress_spring_cli_one_family(populated_compress_context, cli_runner, case_id, caplog):
    """Test to run the decompress command with one family"""
    caplog.set_level(logging.DEBUG)
    # GIVEN a context with a family

    # WHEN running the compress command
    res = cli_runner.invoke(decompress_case, [case_id], obj=populated_compress_context)

    # THEN assert the program exits since no cases where found
    assert res.exit_code == 0
    # THEN assert it was communicated that no families where found
    assert f"Decompressed spring archives in 3 samples" in caplog.text



from functools import partial
from typing import List, Dict, Any

from genologics.entities import Artifact, Project, Container, Containertype, Researcher
from lxml import etree
from lxml.objectify import ElementMaker, ObjectifiedElement

SMP_MAKER = ElementMaker(
    namespace="http://genologics.com/ri/sample",
    nsmap={"smp": "http://genologics.com/ri/sample", "udf": "http://genologics.com/ri/userdefined"},
)
SMP_DETAILS = partial(SMP_MAKER, "details")
SMP_SAMPLECREATION = partial(SMP_MAKER, "samplecreation")

CON_MAKER = ElementMaker(
    namespace="http://genologics.com/ri/container",
    nsmap={"con": "http://genologics.com/ri/container"},
)
CON_DETAILS = partial(CON_MAKER, "details")
CON_CONTAINER = partial(CON_MAKER, "container")

UDF_MAKER = ElementMaker(namespace="http://genologics.com/ri/userdefined", annotate=False)
UDF_FIELD = partial(UDF_MAKER, "field")

ART_MAKER = ElementMaker(
    namespace="http://genologics.com/ri/artifact",
    nsmap={"art": "http://genologics.com/ri/artifact"},
)
ART_DETAILS = partial(ART_MAKER, "details")
ART_ARTIFACT = partial(ART_MAKER, "artifact")

XML = ElementMaker(annotate=False)


def build_container(name: str, con_type: Containertype) -> ObjectifiedElement:
    """Build container in XML."""
    return CON_CONTAINER(XML.name(name), XML.type(uri=con_type.uri))


def build_container_batch(containers: List[ObjectifiedElement]) -> ObjectifiedElement:
    """Build batch with containers."""
    root = CON_DETAILS()
    for container in containers:
        root.append(container)
    return root


def build_sample(
    name: str, project: Project, container: Container, location: str, udfs: Dict[str, Any]
) -> ObjectifiedElement:
    """Build sample in XML."""
    xml_sample = SMP_SAMPLECREATION(
        XML.name(name),
        XML.project(uri=project.uri),
        XML.location(XML.container(uri=container.uri), XML.value(location)),
    )
    for udf_name, udf_value in udfs.items():
        xml_sample.append(UDF_FIELD(udf_value, name=udf_name))
    return xml_sample


def build_sample_batch(samples: List[ObjectifiedElement]) -> ObjectifiedElement:
    """Build batch with samples."""
    root = SMP_DETAILS()
    for sample in samples:
        root.append(sample)
    return root


def build_artifact(artifact: Artifact, reagent_label: str):
    """Build artifact in XML."""
    return ART_ARTIFACT(
        XML("reagent-label", name=reagent_label), XML.name(artifact.name), uri=artifact.uri
    )


def build_artifact_batch(artifacts: List[ObjectifiedElement]) -> ObjectifiedElement:
    """Build bathch with artifacts."""
    root = ART_DETAILS()
    for artifact in artifacts:
        root.append(artifact)
    return root



"""Test FastqHandlerUsalt"""
import re

from cg.apps.microsalt.fastq import FastqHandler


def test_fastq_create(valid_fastq_filename_pattern) -> dict:
    """Test the method that creates balsamic file names for us"""

    # given we have all the necessary inputs to create a filename
    lane = "1"
    flowcell = "FC"
    sample = "ACC123456"
    read = "1"
    undetermined = "u"

    more = {"undetermined": undetermined}

    # when calling the method to create a valid filename
    result_filename = FastqHandler.FastqFileNameCreator.create(
        lane=lane, flowcell=flowcell, sample=sample, read=read, more=more
    )

    # then the filename produced is compatible with the filename rules
    assert re.match(valid_fastq_filename_pattern, result_filename)



"""
This module handles concatenation of balsamic fastq files.

Classes:
    FastqFileNameCreator: Creates valid balsamic filenames
    FastqFileConcatenator: Handles file concatenation
    FastqHandler: Handles fastq file linking
"""
import logging
import os
import shutil
from pathlib import Path
from typing import List

from cg.apps.pipelines.fastqhandler import BaseFastqHandler

LOGGER = logging.getLogger(__name__)


class FastqFileConcatenator:
    """Concatenates a list of files into one"""

    @staticmethod
    def concatenate(files: List, concat_file):
        """Concatenates a list of fastq files"""
        LOGGER.info(FastqFileConcatenator.display_files(files, concat_file))

        with open(concat_file, "wb") as wfd:
            for fil in files:
                with open(fil, "rb") as file_descriptor:
                    shutil.copyfileobj(file_descriptor, wfd)

        size_before = FastqFileConcatenator().size_before(files)
        size_after = FastqFileConcatenator().size_after(concat_file)

        try:
            FastqFileConcatenator().assert_file_sizes(size_before, size_after)
        except AssertionError as error:
            LOGGER.warning(error)

    @staticmethod
    def size_before(files: List):
        """returns the total size of the linked fastq files before concatenation"""

        return sum([os.stat(f).st_size for f in files])

    @staticmethod
    def size_after(concat_file):
        """returns the size of the concatenated fastq files"""

        return os.stat(concat_file).st_size

    @staticmethod
    def assert_file_sizes(size_before, size_after):
        """asserts the file sizes before and after concatenation. If the file sizes differ by more
        than 1 percent, throw an exception"""
        msg = (
            f"Warning: Large file size difference after concatenation!"
            f"Before: {size_before} -> after: {size_after}"
        )

        assert abs(size_before - size_after) / size_before <= 0.01, msg
        LOGGER.info("Concatenation file size check successful!")

    @staticmethod
    def display_files(files: List, concat_file):
        """display file names for logging purposes"""

        concat_file_name = Path(concat_file).name
        msg = (
            f"Concatenating: {', '.join(Path(file_).name for file_ in files)} -> "
            f"{concat_file_name}"
        )

        return msg


class FastqHandler(BaseFastqHandler):
    """Handles fastq file linking"""

    class FastqFileNameCreator(BaseFastqHandler.BaseFastqFileNameCreator):
        """Creates valid balsamic filename from the parameters"""

        @staticmethod
        def create(lane: str, flowcell: str, sample: str, read: str, more: dict = None):
            """Name a FASTQ file following Balsamic conventions. Naming must be
            xxx_R_1.fastq.gz and xxx_R_2.fastq.gz"""
            date = more.get("date", None)
            index = more.get("index", None)
            undetermined = more.get("undetermined", None)

            flowcell = f"{flowcell}-undetermined" if undetermined else flowcell
            date_str = date.strftime("%y%m%d") if date else "171015"
            index = index if index else "XXXXXX"
            return f"{lane}_{date_str}_{flowcell}_{sample}_{index}_R_{read}.fastq.gz"

        @staticmethod
        def get_concatenated_name(linked_fastq_name):
            """"create a name for the concatenated file for some read files"""
            return f"concatenated_{'_'.join(linked_fastq_name.split('_')[-4:])}"

    def __init__(self, config):
        super().__init__(config)
        self.root_dir = config["balsamic"]["root"]

    def link(self, case: str, sample: str, files: List[dict]):
        """Link FASTQ files for a balsamic sample.
        Shall be linked to /<balsamic root directory>/case-id/fastq/"""

        wrk_dir = Path(f"{self.root_dir}/{case}/fastq")

        wrk_dir.mkdir(parents=True, exist_ok=True)

        linked_reads_paths = {1: [], 2: []}
        concatenated_paths = {1: "", 2: ""}

        sorted_files = sorted(files, key=lambda k: k["path"])

        for fastq_data in sorted_files:
            original_fastq_path = Path(fastq_data["path"])
            linked_fastq_name = self.FastqFileNameCreator.create(
                lane=fastq_data["lane"],
                flowcell=fastq_data["flowcell"],
                sample=sample,
                read=fastq_data["read"],
                more={"undetermined": fastq_data["undetermined"]},
            )
            concatenated_fastq_name = self.FastqFileNameCreator.get_concatenated_name(
                linked_fastq_name
            )

            linked_fastq_path = wrk_dir / linked_fastq_name

            linked_reads_paths[fastq_data["read"]].append(linked_fastq_path)
            concatenated_paths[fastq_data["read"]] = f"{wrk_dir}/{concatenated_fastq_name}"

            if not linked_fastq_path.exists():
                LOGGER.info("linking: %s -> %s", original_fastq_path, linked_fastq_path)
                linked_fastq_path.symlink_to(original_fastq_path)
            else:
                LOGGER.debug("destination path already exists: %s", linked_fastq_path)

        LOGGER.info("Concatenation in progress for sample %s.", sample)
        for read in linked_reads_paths:
            FastqFileConcatenator().concatenate(linked_reads_paths[read], concatenated_paths[read])
            self._remove_files(linked_reads_paths[read])

    @staticmethod
    def _remove_files(files):
        for file in files:
            if os.path.isfile(file):
                os.remove(file)

    @staticmethod
    def parse_header(line):
        """Generates a dict with parsed lanes, flowcells and read numbers
        Handle illumina's two different header formats
        @see https://en.wikipedia.org/wiki/FASTQ_format

        @HWUSI-EAS100R:6:73:941:1973#0/1

            HWUSI-EAS100R   the unique instrument name
            6   flowcell lane
            73  tile number within the flowcell lane
            941     'x'-coordinate of the cluster within the tile
            1973    'y'-coordinate of the cluster within the tile
            #0  index number for a multiplexed sample (0 for no indexing)
            /1  the member of a pair, /1 or /2 (paired-end or mate-pair reads only)

        Versions of the Illumina pipeline since 1.4 appear to use #NNNNNN
        instead of #0 for the multiplex ID, where NNNNNN is the sequence of the
        multiplex tag.

        With Casava 1.8 the format of the '@' line has changed:

        @EAS139:136:FC706VJ:2:2104:15343:197393 1:Y:18:ATCACG

            EAS139  the unique instrument name
            136     the run id
            FC706VJ     the flowcell id
            2   flowcell lane
            2104    tile number within the flowcell lane
            15343   'x'-coordinate of the cluster within the tile
            197393  'y'-coordinate of the cluster within the tile
            1   the member of a pair, 1 or 2 (paired-end or mate-pair reads only)
            Y   Y if the read is filtered, N otherwise
            18  0 when none of the control bits are on, otherwise it is an even number
            ATCACG  index sequence
        """

        fastq_meta = {"lane": None, "flowcell": None, "readnumber": None}

        parts = line.split(":")
        if len(parts) == 5:  # @HWUSI-EAS100R:6:73:941:1973#0/1
            fastq_meta["lane"] = parts[1]
            fastq_meta["flowcell"] = "XXXXXX"
            fastq_meta["readnumber"] = parts[-1].split("/")[-1]
        if len(parts) == 10:  # @EAS139:136:FC706VJ:2:2104:15343:197393 1:Y:18:ATCACG
            fastq_meta["lane"] = parts[3]
            fastq_meta["flowcell"] = parts[2]
            fastq_meta["readnumber"] = parts[6].split(" ")[-1]
        if len(parts) == 7:  # @ST-E00201:173:HCLCGALXX:1:2106:22516:34834/1
            fastq_meta["lane"] = parts[3]
            fastq_meta["flowcell"] = parts[2]
            fastq_meta["readnumber"] = parts[-1].split("/")[-1]

        return fastq_meta



"""Tests for reset part of the store API"""
from cg.constants import Pipeline

"""Tests for reset part of the store API"""

from datetime import datetime, timedelta

from cg.store import Store


def test_reset_observation(store: Store):
    # GIVEN a store with a case with loqus-links
    family = add_family(store)
    sample = add_sample(store, loqusdb_links=True)
    store.relate_sample(family=family, sample=sample, status="unknown")
    assert sample.loqusdb_id is not None

    # WHEN calling reset observations
    store.reset_observations(case_id=family.internal_id)

    # THEN the links to observations in loqusdb should have been reset
    assert sample.loqusdb_id is None


def ensure_application_version(disk_store, application_tag="dummy_tag"):
    """utility function to return existing or create application version for tests"""
    application = disk_store.application(tag=application_tag)
    if not application:
        application = disk_store.add_application(
            tag=application_tag,
            category="wgs",
            description="dummy_description",
            percent_kth=80,
        )
        disk_store.add_commit(application)

    prices = {"standard": 10, "priority": 20, "express": 30, "research": 5}
    version = disk_store.application_version(application, 1)
    if not version:
        version = disk_store.add_version(application, 1, valid_from=datetime.now(), prices=prices)

        disk_store.add_commit(version)
    return version


def ensure_customer(disk_store, customer_id="cust_test"):
    """utility function to return existing or create customer for tests"""
    customer_group = disk_store.customer_group("dummy_group")
    if not customer_group:
        customer_group = disk_store.add_customer_group("dummy_group", "dummy group")

        customer = disk_store.add_customer(
            internal_id=customer_id,
            name="Test Customer",
            scout_access=False,
            customer_group=customer_group,
            invoice_address="dummy_address",
            invoice_reference="dummy_reference",
        )
        disk_store.add_commit(customer)
    customer = disk_store.customer(customer_id)
    return customer


def add_sample(store, sample_name="sample_test", loqusdb_links=False):
    """utility function to add a sample to use in tests"""

    customer = ensure_customer(store)
    application_version_id = ensure_application_version(store).id
    sample = store.add_sample(name=sample_name, sex="unknown")
    sample.application_version_id = application_version_id
    sample.customer = customer
    if loqusdb_links:
        sample.loqusdb_id = True
    store.add_commit(sample)
    return sample


def ensure_panel(disk_store, panel_id="panel_test", customer_id="cust_test"):
    """utility function to add a panel to use in tests"""
    customer = ensure_customer(disk_store, customer_id)
    panel = disk_store.panel(panel_id)
    if not panel:
        panel = disk_store.add_panel(
            customer=customer,
            name=panel_id,
            abbrev=panel_id,
            version=1.0,
            date=datetime.now(),
            genes=1,
        )
        disk_store.add_commit(panel)
    return panel


def add_family(
    disk_store,
    family_id="family_test",
    customer_id="cust_test",
    ordered_days_ago=0,
    action=None,
    priority=None,
    data_analysis=Pipeline.MIP_DNA,
):
    """utility function to add a family to use in tests"""
    panel = ensure_panel(disk_store)
    customer = ensure_customer(disk_store, customer_id)
    family = disk_store.add_family(data_analysis=data_analysis, name=family_id, panels=panel.name)
    family.customer = customer
    family.ordered_at = datetime.now() - timedelta(days=ordered_days_ago)
    if action:
        family.action = action
    if priority:
        family.priority = priority
    disk_store.add_commit(family)
    return family



"""Functions to get cases from the store database"""

from alchy import Query
import datetime as dt

from cg.store import models
from cg.store.api import Store


def ready_for_spring_compression(store: Store, date_threshold: dt.datetime) -> Query:
    """Fetch all cases that are ready to be compressed by SPRING"""
    cases = (
        store.query(models.Family)
        .filter(models.Family.action.isnot("running"), models.Family.action.isnot("analyze"))
        .filter(models.Family.created_at < date_threshold)
        .order_by(models.Family.created_at.asc())
    )

    return cases



from cg.constants import Pipeline
from cg.store import Store
from cg.meta.workflow.mip import MipAnalysisAPI
import logging


def test_config(analysis_store: Store, analysis_api: MipAnalysisAPI):

    # GIVEN a status db with a family
    family_obj = analysis_store.families().first()
    assert family_obj is not None
    for link_obj in family_obj.links:
        link_obj.sample.application_version.application.min_sequencing_depth = 10
        analysis_store.commit()

    # WHEN generating the MIP config for the family
    mip_config = analysis_api.pedigree_config(family_obj, pipeline=Pipeline.MIP_DNA)

    # THEN it should fill in values accordingly
    assert len(mip_config["samples"]) == len(family_obj.links)


def test_get_latest_data_genome_build(analysis_api: MipAnalysisAPI, case_id: str):

    # GIVEN

    # WHEN
    trending_data = analysis_api.get_latest_metadata(case_id)

    # THEN contains genome_build
    assert trending_data["genome_build"]


def test_get_latest_data_rank_model_version(analysis_api: MipAnalysisAPI, case_id: str):
    # GIVEN

    # WHEN
    trending_data = analysis_api.get_latest_metadata(case_id)

    # THEN contains rankmodelversion
    assert trending_data["rank_model_version"]


def test_get_latest_metadata_logging(caplog, analysis_api: MipAnalysisAPI):
    # GIVEN an initialised report_api and the deliver_api does not have what we want
    family_id = "case_missing_data"
    # WHEN failing to get latest trending data for a family
    latest_data = analysis_api.get_latest_metadata(family_id)

    with caplog.at_level(logging.WARNING):
        assert family_id in caplog.text
    # THEN there should be a log entry about this
    assert not latest_data



"""Class to hold information about scout load config"""

from typing import Optional, List
from typing_extensions import Literal
from datetime import datetime
from pydantic import BaseModel


class ScoutIndividual(BaseModel):
    sample_id: str
    father: Optional[str] = None
    mother: Optional[str] = None
    display_name: Optional[str] = None
    sex: Literal["male", "female", "unknown"]
    phenotype: Literal["affected", "unaffected", "unknown"]
    capture_kit: Optional[str] = None
    alignment_path: Optional[str] = None
    chromograph: Optional[str] = None
    rhocall_bed: Optional[str] = None
    rhocall_wig: Optional[str] = None
    tiddit_coverage_wig: Optional[str] = None
    upd_regions_bed: Optional[str] = None
    upd_sites_bed: Optional[str] = None
    vcf2cytosure: Optional[str] = None
    analysis_type: Literal["wgs", "wes", "mixed", "unknown", "panel", "external"]
    confirmed_sex: Optional[bool] = None
    confirmed_parent: Optional[bool] = None
    is_sma: Optional[bool] = None
    is_sma_carrier: Optional[bool] = None
    smn1_cn: Optional[int] = None  # CopyNumber
    smn2_cn: Optional[int] = None  # CopyNumber
    smn2delta78_cn: Optional[int] = None  # CopyNumber
    smn_27134_cn: Optional[int] = None  # CopyNumber
    tumor_type: Optional[str] = None
    tmb: Optional[str] = None
    msi: Optional[str] = None
    tumor_purity: Optional[float] = None
    tissue_type: Optional[str] = None


class ScoutLoadConfig(BaseModel):
    owner: str
    family: str
    family_name: Optional[str] = None
    synopsis: Optional[str] = None
    phenotype_terms: Optional[List[str]] = None
    gene_panels: Optional[List[str]] = None
    default_gene_panels: List[str] = []
    cohorts: Optional[List[str]] = None
    human_genome_build: str
    rank_model_version: Optional[str] = None
    rank_score_threshold: int
    sv_rank_model_version: Optional[str] = None
    analysis_date: datetime
    samples: List[ScoutIndividual]
    vcf_snv_research: Optional[str] = None
    vcf_snv: Optional[str] = None
    vcf_sv: Optional[str] = None
    vcf_str: Optional[str] = None
    vcf_cancer: Optional[str] = None
    vcf_cancer_sv: Optional[str] = None
    vcf_sv_research: Optional[str] = None
    vcf_cancer_research: Optional[str] = None
    vcf_cancer_sv_research: Optional[str] = None
    smn_tsv: Optional[str] = None
    peddy_ped: Optional[str] = None
    peddy_sex: Optional[str] = None
    peddy_check: Optional[str] = None
    delivery_report: Optional[str] = None
    cnv_report: Optional[str] = None
    multiqc: Optional[str] = None
    track: Literal["rare", "cancer"] = "rare"
    chromograph_image_files: Optional[List[str]]
    chromograph_prefixes: Optional[List[str]]



"""Set data in the status database and LIMS"""
import datetime
import getpass
import logging

import click
from cg.apps.lims import LimsAPI
from .families import families
from .family import family
from cg.constants import FLOWCELL_STATUS
from cg.exc import LimsDataError
from cg.store import Store, models

CONFIRM = "Continue?"
HELP_KEY_VALUE = "Give a property on sample and the value to set it to, e.g. -kv name Prov52"
HELP_SKIP_LIMS = "Skip setting value in LIMS"
HELP_YES = "Answer yes on all confirmations"
OPTION_LONG_KEY_VALUE = "--key-value"
OPTION_LONG_SKIP_LIMS = "--skip-lims"
OPTION_LONG_YES = "--yes"
OPTION_SHORT_KEY_VALUE = "-kv"
OPTION_SHORT_YES = "-y"
NOT_CHANGEABLE_SAMPLE_ATTRIBUTES = [
    "application_version_id",
    "customer_id",
    "deliveries",
    "flowcells",
    "id",
    "internal_id",
    "invoice",
    "invoice_id",
    "is_external",
    "links",
    "state",
    "to_dict",
]

LOG = logging.getLogger(__name__)


@click.group("set")
@click.pass_context
def set_cmd(context):
    """Update information in the database."""
    context.obj["status_db"] = Store(context.obj["database"])
    context.obj["lims_api"] = LimsAPI(context.obj)


@set_cmd.command()
@click.option(
    "-id",
    "--identifier",
    "identifiers",
    nargs=2,
    type=click.Tuple([str, str]),
    multiple=True,
    help="Give an identifier on sample and the value to use it with, e.g. -id name Prov52",
)
@click.option(
    "-kv",
    "--key-value",
    "kwargs",
    nargs=2,
    type=click.Tuple([str, str]),
    multiple=True,
    help="Give a property on sample and the value to set it to, e.g. -kv name Prov52",
)
@click.option("--skip-lims", is_flag=True, help="Skip setting value in LIMS")
@click.option("-y", "--yes", is_flag=True, help="Answer yes on all confirmations")
@click.argument("case_id", required=False)
@click.pass_context
def samples(
    context: click.Context,
    identifiers: click.Tuple([str, str]),
    kwargs: click.Tuple([str, str]),
    skip_lims: bool,
    yes: bool,
    case_id: str,
):
    """Set values on many samples at the same time"""
    store = context.obj["status_db"]
    sample_objs = _get_samples(case_id, identifiers, store)

    if not sample_objs:
        LOG.error("No samples to alter!")
        context.abort()

    LOG.info("Would alter samples:")

    for sample_obj in sample_objs:
        LOG.info(f"{sample_obj}")

    if not (yes or click.confirm(CONFIRM)):
        context.abort()

    for sample_obj in sample_objs:
        context.invoke(
            sample, sample_id=sample_obj.internal_id, kwargs=kwargs, yes=yes, skip_lims=skip_lims
        )


def _get_samples(
    case_id: str, identifiers: click.Tuple([str, str]), store: Store
) -> [models.Sample]:
    """Get samples that match both case_id and identifiers if given"""
    samples_by_case_id = None
    samples_by_id = None

    if case_id:
        samples_by_case_id = _get_samples_by_case_id(case_id, store)

    if identifiers:
        samples_by_id = _get_samples_by_identifiers(identifiers, store)

    if case_id and identifiers:
        sample_objs = set(set(samples_by_case_id) & set(samples_by_id))
    else:
        sample_objs = samples_by_case_id or samples_by_id

    return sample_objs


def _get_samples_by_identifiers(
    identifiers: click.Tuple([str, str]), store: Store
) -> models.Sample:
    """Get samples matched by given set of identifiers"""
    identifier_args = {}
    for identifier_name, identifier_value in identifiers:
        identifier_args[identifier_name] = identifier_value
    return store.samples_by_ids(**identifier_args)


def _get_samples_by_case_id(case_id: str, store: Store) -> [models.Sample]:
    """Get samples on a given case-id"""
    case = store.family(internal_id=case_id)
    return [link.sample for link in case.links] if case else []


def is_locked_attribute_on_sample(key, skip_attributes):
    """Returns true if the attribute is private or in the skip list"""
    return is_private_attribute(key) or key in skip_attributes


def is_private_attribute(key):
    """Returns true if key has name indicating it is private"""
    return key.startswith("_")


def list_changeable_sample_attributes(sample_obj: models.Sample = None, skip_attributes: [] = None):
    """List changeable attributes on sample and its current value"""

    sample_attributes = models.Sample.__dict__.keys()
    for attribute in sample_attributes:
        if is_locked_attribute_on_sample(attribute, skip_attributes):
            continue

        message = attribute

        if sample_obj:
            message += f": {sample_obj.__dict__.get(attribute)}"

        LOG.info(message)


def show_set_sample_help(sample_obj: models.Sample = "None") -> None:
    """Show help for the set sample command"""
    LOG.info("sample_id: optional, internal_id of sample to set value on")
    show_option_help(long_name=OPTION_LONG_SKIP_LIMS, help_text=HELP_SKIP_LIMS)
    show_option_help(short_name=OPTION_SHORT_YES, long_name=OPTION_LONG_YES, help_text=HELP_YES)
    show_option_help(
        short_name=OPTION_SHORT_KEY_VALUE, long_name=OPTION_LONG_KEY_VALUE, help_text=HELP_KEY_VALUE
    )
    list_changeable_sample_attributes(sample_obj, skip_attributes=NOT_CHANGEABLE_SAMPLE_ATTRIBUTES)
    LOG.info(f"To set apptag use '{OPTION_SHORT_KEY_VALUE} application_version [APPTAG]")
    LOG.info(f"To set customer use '{OPTION_SHORT_KEY_VALUE} customer [CUSTOMER]")
    LOG.info(
        f"To set priority use '{OPTION_SHORT_KEY_VALUE} priority [priority as text or " f"number]"
    )


def show_option_help(short_name: str = None, long_name: str = None, help_text: str = None):
    """Show help for one option"""
    help_message = f"Use "

    if short_name:
        help_message += f"'{short_name}'"

    if short_name and long_name:
        help_message += " or "

    if long_name:
        help_message += f"'{long_name}'"

    if help_text:
        help_message += f": {help_text}"

    LOG.info(help_message)


@set_cmd.command()
@click.argument("sample_id", required=False)
@click.option(
    OPTION_SHORT_KEY_VALUE,
    OPTION_LONG_KEY_VALUE,
    "kwargs",
    nargs=2,
    type=click.Tuple([str, str]),
    multiple=True,
    help=HELP_KEY_VALUE,
)
@click.option(OPTION_LONG_SKIP_LIMS, is_flag=True, help=HELP_SKIP_LIMS)
@click.option(OPTION_SHORT_YES, OPTION_LONG_YES, is_flag=True, help=HELP_YES)
@click.option("--help", is_flag=True)
@click.pass_context
def sample(context, sample_id, kwargs, skip_lims, yes, help):
    sample_obj = context.obj["status_db"].sample(internal_id=sample_id)

    if help:
        show_set_sample_help(sample_obj)

    if sample_obj is None:
        LOG.error(f"Can't find sample {sample_id}")
        context.abort()

    for key, value in kwargs:

        if is_locked_attribute_on_sample(key, NOT_CHANGEABLE_SAMPLE_ATTRIBUTES):
            LOG.warning(f"{key} is not a changeable attribute on sample")
            continue
        if not hasattr(sample_obj, key):
            LOG.warning(f"{key} is not a property of sample")
            continue

        new_key = key
        new_value = value

        if key in ["customer", "application_version", "priority"]:
            if key == "priority":
                if isinstance(value, str) and not value.isdigit():
                    new_key = "priority_human"
            elif key == "customer":
                new_value = context.obj["status_db"].customer(value)
            elif key == "application_version":
                new_value = context.obj["status_db"].current_application_version(value)

            if not new_value:
                LOG.error(f"{key} {value} not found, aborting")
                context.abort()

        old_value = getattr(sample_obj, new_key)

        LOG.info(
            f"Would change from {new_key}={old_value} to {new_key}={new_value} on {sample_obj}"
        )

        if not (yes or click.confirm(CONFIRM)):
            continue

        setattr(sample_obj, new_key, new_value)
        _update_comment(_generate_comment(new_key, old_value, new_value), sample_obj)
        context.obj["status_db"].commit()

    if not skip_lims:

        for key, value in kwargs:

            if key == "application_version":
                new_key = "application"
            else:
                new_key = key

            if key == "priority":
                new_value = sample_obj.priority_human
            else:
                new_value = value

            LOG.info(f"Would set {new_key} to {new_value} for {sample_obj.internal_id} in LIMS")

            if not (yes or click.confirm(CONFIRM)):
                context.abort()

            try:
                context.obj["lims_api"].update_sample(lims_id=sample_id, **{new_key: new_value})
                LOG.info(f"Set LIMS/{new_key} to {new_value}")
            except LimsDataError as err:
                LOG.error(f"Failed to set LIMS/{new_key} to {new_value}, {err.message}")


def _generate_comment(what, old_value, new_value):
    """Generate a comment that can be used in the comment field to describe updated value"""
    return f"\n{what} changed from {str(old_value)} to {str(new_value)}."


def _update_comment(comment, obj):
    """Appends the comment on obj including a timestamp"""
    if comment:
        timestamp = str(datetime.datetime.now())[:-10]
        if obj.comment is None:
            obj.comment = f"{timestamp}-{getpass.getuser()}: {comment}"
        else:
            obj.comment = f"{timestamp}-{getpass.getuser()}: {comment}" + "\n" + obj.comment


@set_cmd.command()
@click.option("-s", "--status", type=click.Choice(FLOWCELL_STATUS))
@click.argument("flowcell_name")
@click.pass_context
def flowcell(context, flowcell_name, status):
    """Update information about a flowcell"""
    flowcell_obj = context.obj["status_db"].flowcell(flowcell_name)

    if flowcell_obj is None:
        LOG.warning(f"flowcell not found: {flowcell_name}")
        context.abort()
    prev_status = flowcell_obj.status
    flowcell_obj.status = status

    context.obj["status_db"].commit()
    LOG.info(f"{flowcell_name} set: {prev_status} -> {status}")


set_cmd.add_command(family)
set_cmd.add_command(families)



"""Test FastqHandler"""
import os
from pathlib import Path

from cg.apps.balsamic.fastq import FastqHandler


def test_link_file_count(tmpdir, cg_config, link_family, link_sample, simple_files_data):
    """Test method to test that the right number of files are created by linking"""

    # given some fastq-files belonging to family and sample
    link_files = simple_files_data
    assert len(tmpdir.listdir()) == len(link_files)
    link_dir = Path(f"{tmpdir}/{link_family}/fastq")
    assert not os.path.exists(link_dir)

    # when calling the method to link
    FastqHandler(cg_config).link(case=link_family, sample=link_sample, files=link_files)

    # then the linking should have created on directory for the linked files
    assert os.path.exists(link_dir)
    assert len(tmpdir.listdir()) == len(link_files) + 1

    # then we should have a new directory with one concatenated file per read direction
    assert (
        len([name for name in os.listdir(link_dir) if os.path.isfile(os.path.join(link_dir, name))])
        == 2
    )


def test_link_file_content(
    tmpdir, cg_config, link_family, link_sample, simple_files_data, content_r1, content_r2
):
    """Test method to test that balsamic files are linked properly"""

    # given some fastq-files belonging to family and sample
    link_files = simple_files_data
    link_dir = Path(f"{tmpdir}/{link_family}/fastq")

    # when calling the method to link
    FastqHandler(cg_config).link(case=link_family, sample=link_sample, files=link_files)

    # then the first concatenated file should contain 'ABCD' and the other 'DEFG'
    linked_files = [
        name for name in os.listdir(link_dir) if os.path.isfile(os.path.join(link_dir, name))
    ]

    file_contents = []

    for file_name in linked_files:
        file_path = os.path.join(link_dir, file_name)

        with open(file_path, "rt") as file:
            file_contents.append(file.read())

    assert content_r1 in file_contents
    assert content_r2 in file_contents


def test_link_file_content_reversed(
    tmpdir, cg_config, link_family, link_sample, simple_files_data_reversed, content_r1, content_r2
):
    """Test method to test that balsamic files are linked properly"""

    # given some fastq-files belonging to family and sample
    link_files = simple_files_data_reversed
    link_dir = Path(f"{tmpdir}/{link_family}/fastq")

    # when calling the method to link
    FastqHandler(cg_config).link(case=link_family, sample=link_sample, files=link_files)

    # then the first concatenated file should contain 'ABCD' and the other 'DEFG'
    linked_files = [
        name for name in os.listdir(link_dir) if os.path.isfile(os.path.join(link_dir, name))
    ]

    file_contents = []

    for file_name in linked_files:
        file_path = os.path.join(link_dir, file_name)

        with open(file_path, "rt") as file:
            file_contents.append(file.read())

    assert content_r1 in file_contents
    assert content_r2 in file_contents



"""CLI support to start microsalt"""

import json
import logging
from pathlib import Path

import click

from cg.apps.hk import HousekeeperAPI
from cg.apps.lims import LimsAPI
from cg.apps.microsalt.fastq import FastqHandler
from cg.cli.workflow.microsalt.store import store as store_cmd
from cg.meta.workflow.microsalt import MicrosaltAnalysisAPI
from cg.store import Store
from cg.utils.commands import Process

LOG = logging.getLogger(__name__)


@click.group("microsalt", invoke_without_command=True)
@click.option("-d", "--dry-run", is_flag=True, help="print to console")
@click.option("-t", "--ticket", help="include all microbial samples for a ticket")
@click.pass_context
def microsalt(context: click.Context, ticket: str, dry_run: bool):
    """Microbial workflow"""
    context.obj["status_db"] = Store(context.obj["database"])
    hk_api = HousekeeperAPI(context.obj)
    lims_api = LimsAPI(context.obj)
    analysis_api = MicrosaltAnalysisAPI(
        db=context.obj["status_db"],
        hk_api=hk_api,
        lims_api=lims_api,
        fastq_handler=FastqHandler(context.obj),
        config=context.obj["microsalt"],
    )
    context.obj["microsalt_analysis_api"] = analysis_api

    if context.invoked_subcommand:
        return

    if not ticket:
        LOG.error("Please provide a ticket")
        context.abort()

    if not analysis_api.has_flowcells_on_disk(ticket=ticket):
        LOG.warning(
            f"Ticket {ticket} not ready to run, requesting flowcells from long term storage!",
        )
        analysis_api.request_removed_flowcells(ticket=ticket)
        return

    # execute the analysis!
    context.invoke(config_case, ticket=ticket, dry_run=dry_run)
    context.invoke(link, ticket=ticket, dry_run=dry_run)
    context.invoke(run, ticket=ticket, dry_run=dry_run)


@click.command()
@click.option("-d", "--dry-run", is_flag=True, help="only print to console")
@click.option("-t", "--ticket", help="link all microbial samples for a ticket")
@click.argument("sample_id", required=False)
@click.pass_context
def link(context: click.Context, dry_run: bool, ticket: str, sample_id: str):
    """Link microbial FASTQ files for a SAMPLE_ID"""

    analysis_api = context.obj["microsalt_analysis_api"]

    if not any([ticket, sample_id]):
        LOG.error("provide ticket and/or sample")
        context.abort()

    if dry_run:
        return

    analysis_api.link_samples(ticket=ticket, sample_id=sample_id)


@click.command("config-case")
@click.option("-d", "--dry-run", is_flag=True, help="print config to console")
@click.option("-t", "--ticket", help="create config-case all microbial samples for an order")
@click.argument("sample_id", required=False)
@click.pass_context
def config_case(context: click.Context, dry_run: bool, ticket: int, sample_id: str):
    """ Create a config file on case level for microSALT """

    if not ticket and not sample_id:
        LOG.error("Provide ticket and/or sample")
        context.abort()

    sample_objs = context.obj["microsalt_analysis_api"].get_samples(ticket, sample_id)

    if not sample_objs:
        LOG.error("No sample found for that ticket/sample_id")
        context.abort()

    parameters = [
        context.obj["microsalt_analysis_api"].get_parameters(sample_obj)
        for sample_obj in sample_objs
    ]

    filename = str(ticket) if ticket else sample_id
    outfilename = (Path(context.obj["microsalt"]["queries_path"]) / filename).with_suffix(".json")
    if dry_run:
        print(json.dumps(parameters, indent=4, sort_keys=True))
        return

    with open(outfilename, "w") as outfile:
        json.dump(parameters, outfile, indent=4, sort_keys=True)


@click.command()
@click.option("-d", "--dry-run", "dry_run", is_flag=True, help="print command to console")
@click.option(
    "-c",
    "--config-case",
    "config_case_path",
    required=False,
    type=click.Path(exists=True, file_okay=True, dir_okay=False, resolve_path=True),
    help="optionally change the config-case",
)
@click.argument("ticket", type=int)
@click.pass_context
def run(context: click.Context, dry_run: bool, config_case_path: click.Path, ticket: int):
    """ Start microSALT with an order_id """

    microbial_api = context.obj["microsalt_analysis_api"]

    microsalt_bin = context.obj["microsalt"]["binary_path"]
    microsalt_env = context.obj["microsalt"]["conda_env"]
    fastq_path = Path(context.obj["microsalt"]["root"]) / "fastq" / str(ticket)

    if config_case_path:
        config_case_path = Path(config_case_path)
    else:
        queries_path = Path(context.obj["microsalt"]["queries_path"])
        config_case_path = (queries_path / str(ticket)).with_suffix(".json")

    process = Process(binary=microsalt_bin, environment=microsalt_env)
    analyse_command = [
        "analyse",
        str(config_case_path.absolute()),
        "--input",
        str(fastq_path.absolute()),
    ]
    process.run_command(parameters=analyse_command, dry_run=dry_run)

    if not dry_run:
        microbial_api.set_statusdb_action(name=ticket, action="running")


microsalt.add_command(store_cmd)
microsalt.add_command(run)
microsalt.add_command(config_case)
microsalt.add_command(link)



import logging
from typing import List

LOG = logging.getLogger(__name__)
SEX_MAP = {"male": "M", "female": "F"}


class LimsHandler:
    @staticmethod
    def to_lims(customer: str, samples: List[dict]) -> List[dict]:
        """Convert order input to lims interface input."""
        samples_lims = []
        for sample in samples:
            LOG.debug(f"{sample['name']}: prepare LIMS input")
            samples_lims.append(
                {
                    "name": sample["name"],
                    "container": sample.get("container") or "Tube",
                    "container_name": sample.get("container_name"),
                    "well_position": sample.get("well_position"),
                    "index_sequence": sample.get("index_sequence"),
                    "udfs": {
                        "application": sample["application"],
                        "capture_kit": sample.get("capture_kit"),
                        "comment": sample.get("comment"),
                        "concentration": sample.get("concentration"),
                        "concentration_weight": sample.get("concentration_weight"),
                        "customer": customer,
                        "data_analysis": sample.get("data_analysis"),
                        "elution_buffer": sample.get("elution_buffer"),
                        "extraction_method": sample.get("extraction_method"),
                        "family_name": sample.get("family_name") or "NA",
                        "formalin_fixation_time": sample.get("formalin_fixation_time"),
                        "index": sample.get("index"),
                        "index_number": sample.get("index_number"),
                        "organism": sample.get("organism"),
                        "organism_other": sample.get("organism_other"),
                        "pool": sample.get("pool"),
                        "post_formalin_fixation_time": sample.get("post_formalin_fixation_time"),
                        "priority": sample.get("priority") or "standard",
                        "quantity": sample.get("quantity"),
                        "reference_genome": sample.get("reference_genome"),
                        "require_qcok": sample.get("require_qcok") or False,
                        "rml_plate_name": sample.get("rml_plate_name"),
                        "sex": SEX_MAP.get(sample.get("sex"), "unknown"),
                        "source": sample.get("source") or "NA",
                        "tissue_block_size": sample.get("tissue_block_size"),
                        "tumour": sample.get("tumour") or False,
                        "tumour_purity": sample.get("tumour_purity"),
                        "volume": sample.get("volume"),
                        "well_position_rml": sample.get("well_position_rml"),
                        "verified_organism": sample.get("verified_organism"),
                    },
                }
            )
        return samples_lims

    def process_lims(self, data: dict, samples: List[dict]):
        """Process samples to add them to LIMS."""
        samples_lims = self.to_lims(data["customer"], samples)
        project_name = data["ticket"] or data["name"]
        project_data = self.lims.submit_project(project_name, samples_lims)
        lims_map = self.lims.get_samples(projectlimsid=project_data["id"], map_ids=True)
        return project_data, lims_map
